"""ORM classes for the Matchbox PostgreSQL database."""

import json
from typing import Any, Literal

from sqlalchemy import (
    BIGINT,
    BOOLEAN,
    INTEGER,
    SMALLINT,
    CheckConstraint,
    Column,
    DateTime,
    Enum,
    ForeignKey,
    Identity,
    Index,
    UniqueConstraint,
    select,
    text,
    update,
)
from sqlalchemy.dialects.postgresql import BYTEA, JSONB, TEXT, insert
from sqlalchemy.orm import Session, relationship, selectinload

from matchbox.common.dtos import Collection as CommonCollection
from matchbox.common.dtos import (
    CollectionName,
    LocationConfig,
    ModelType,
    ResolutionName,
    ResolutionPath,
    ResolutionType,
    RunID,
    UploadStage,
)
from matchbox.common.dtos import ModelConfig as CommonModelConfig
from matchbox.common.dtos import Resolution as CommonResolution
from matchbox.common.dtos import Run as CommonRun
from matchbox.common.dtos import SourceConfig as CommonSourceConfig
from matchbox.common.dtos import SourceField as CommonSourceField
from matchbox.common.exceptions import (
    MatchboxCollectionNotFoundError,
    MatchboxResolutionAlreadyExists,
    MatchboxResolutionNotFoundError,
    MatchboxRunNotFoundError,
)
from matchbox.server.postgresql.db import MBDB
from matchbox.server.postgresql.mixin import CountMixin


class Collections(CountMixin, MBDB.MatchboxBase):
    """Named collections of resolutions and runs."""

    __tablename__ = "collections"

    collection_id = Column(BIGINT, primary_key=True, autoincrement=True)
    name = Column(TEXT, nullable=False)

    # Relationships
    runs = relationship("Runs", back_populates="collection")

    # Constraints
    __table_args__ = (UniqueConstraint("name", name="collections_name_key"),)

    @classmethod
    def from_name(
        cls,
        name: CollectionName,
        session: Session | None = None,
    ) -> "Collections":
        """Resolve a collection name to a Collections object.

        Args:
            name: The name of the collection to resolve.
            session: Optional session to use for the query.

        Raises:
            MatchboxCollectionNotFoundError: If the collection doesn't exist.
        """
        query = select(cls).where(cls.name == name).options(selectinload(cls.runs))

        if session:
            collection = session.execute(query).scalar_one_or_none()
        else:
            with MBDB.get_session() as session:
                collection = session.execute(query).scalar_one_or_none()

        if not collection:
            raise MatchboxCollectionNotFoundError(f"Collection '{name}' not found.")

        return collection

    def to_dto(self) -> CommonCollection:
        """Convert ORM collection to a matchbox.common Collection object."""
        run_ids: list[RunID] = []
        default_run = None
        if runs := self.runs:
            run_ids = [r.run_id for r in runs]
            default_run_list = [r.run_id for r in runs if r.is_default]
            if default_run_list:
                default_run = default_run_list[0]

        return CommonCollection(runs=run_ids, default_run=default_run)


class Runs(CountMixin, MBDB.MatchboxBase):
    """Runs of collections of resolutions."""

    __tablename__ = "runs"

    run_id = Column(BIGINT, primary_key=True, autoincrement=True)
    collection_id = Column(
        BIGINT,
        ForeignKey("collections.collection_id", ondelete="CASCADE"),
        nullable=False,
    )
    is_mutable = Column(BOOLEAN, default=False)
    is_default = Column(BOOLEAN, default=False)

    # Relationships
    collection = relationship("Collections", back_populates="runs")
    resolutions = relationship("Resolutions", back_populates="run")

    # Constraints
    __table_args__ = (
        UniqueConstraint("collection_id", "run_id", name="unique_run_id"),
        Index(
            "ix_default_run_collection",
            "collection_id",
            unique=True,
            postgresql_where=text("is_default = true"),
        ),
    )

    @classmethod
    def from_id(
        cls,
        collection: CollectionName,
        run_id: RunID,
        session: Session | None = None,
    ) -> "Runs":
        """Resolve a collection and run name to a Runs object.

        Args:
            collection: The name of the collection containing the run.
            run_id: The ID of the run within that collection.
            session: Optional session to use for the query.

        Raises:
            MatchboxRunNotFoundError: If the run doesn't exist.
        """
        query = (
            select(cls)
            .where(cls.run_id == run_id)
            .options(
                selectinload(cls.resolutions)
                .selectinload(Resolutions.source_config)
                .selectinload(SourceConfigs.fields),
                selectinload(cls.resolutions).selectinload(Resolutions.model_config),
                selectinload(cls.collection),
            )
        )

        if session:
            run_orm = session.execute(query).scalar_one_or_none()
        else:
            with MBDB.get_session() as session:
                run_orm = session.execute(query).scalar_one_or_none()

        if not run_orm:
            raise MatchboxRunNotFoundError

        if run_orm.collection.name != collection:
            raise MatchboxRunNotFoundError(
                run_id=id,
                message=f"Run {id} not found in collection {collection}",
            )

        return run_orm

    def to_dto(self) -> CommonRun:
        """Convert ORM run to a matchbox.common Run object."""
        resolutions: dict[ResolutionName, CommonResolution] = {}
        if self.resolutions:
            resolutions = {
                resolution.name: resolution.to_dto() for resolution in self.resolutions
            }

        return CommonRun(
            run_id=self.run_id,
            is_default=self.is_default,
            is_mutable=self.is_mutable,
            resolutions=resolutions,
        )


class ResolutionFrom(CountMixin, MBDB.MatchboxBase):
    """Resolution lineage closure table with cached truth values."""

    __tablename__ = "resolution_from"

    # Columns
    parent = Column(
        BIGINT,
        ForeignKey("resolutions.resolution_id", ondelete="CASCADE"),
        primary_key=True,
    )
    child = Column(
        BIGINT,
        ForeignKey("resolutions.resolution_id", ondelete="CASCADE"),
        primary_key=True,
    )
    level = Column(INTEGER, nullable=False)
    truth_cache = Column(SMALLINT, nullable=True)

    # Constraints
    __table_args__ = (
        CheckConstraint("parent != child", name="no_self_reference"),
        CheckConstraint("level > 0", name="positive_level"),
    )


class Resolutions(CountMixin, MBDB.MatchboxBase):
    """Table of resolution points corresponding to models, and sources.

    Resolutions produce probabilities or own data in the clusters table.
    """

    __tablename__ = "resolutions"

    # Columns
    resolution_id = Column(BIGINT, primary_key=True, autoincrement=True)
    run_id = Column(
        BIGINT, ForeignKey("runs.run_id", ondelete="CASCADE"), nullable=False
    )
    upload_stage = Column(
        Enum(UploadStage, native_enum=True, name="upload_stages", schema="mb"),
        nullable=False,
        default=UploadStage.READY,
    )
    name = Column(TEXT, nullable=False)
    description = Column(TEXT, nullable=True)
    type = Column(TEXT, nullable=False)
    fingerprint = Column(BYTEA, nullable=False)
    truth = Column(SMALLINT, nullable=True)

    # Relationships
    source_config = relationship(
        "SourceConfigs", back_populates="source_resolution", uselist=False
    )
    model_config = relationship(
        "ModelConfigs", back_populates="model_resolution", uselist=False
    )
    probabilities = relationship(
        "Probabilities",
        back_populates="proposed_by",
        passive_deletes=True,
    )
    results = relationship(
        "Results",
        back_populates="proposed_by",
        passive_deletes=True,
    )
    children = relationship(
        "Resolutions",
        secondary=ResolutionFrom.__table__,
        primaryjoin="Resolutions.resolution_id == ResolutionFrom.parent",
        secondaryjoin="Resolutions.resolution_id == ResolutionFrom.child",
        backref="parents",
    )
    run = relationship("Runs", back_populates="resolutions")

    # Constraints
    __table_args__ = (
        CheckConstraint(
            "type IN ('model', 'source')",
            name="resolution_type_constraints",
        ),
        UniqueConstraint("run_id", "name", name="resolutions_name_key"),
    )

    @property
    def ancestors(self) -> set["Resolutions"]:
        """Returns all ancestors (parents, grandparents, etc.) of this resolution."""
        with MBDB.get_session() as session:
            ancestor_query = (
                select(Resolutions)
                .select_from(Resolutions)
                .join(
                    ResolutionFrom, Resolutions.resolution_id == ResolutionFrom.parent
                )
                .where(ResolutionFrom.child == self.resolution_id)
            )
            return set(session.execute(ancestor_query).scalars().all())

    @property
    def descendants(self) -> set["Resolutions"]:
        """Returns descendants (children, grandchildren, etc.) of this resolution."""
        with MBDB.get_session() as session:
            descendant_query = (
                select(Resolutions)
                .select_from(Resolutions)
                .join(ResolutionFrom, Resolutions.resolution_id == ResolutionFrom.child)
                .where(ResolutionFrom.parent == self.resolution_id)
            )
            return set(session.execute(descendant_query).scalars().all())

    def get_lineage(
        self, sources: list["SourceConfigs"] | None = None, threshold: int | None = None
    ) -> list[tuple[int, int, float | None]]:
        """Returns lineage ordered by priority.

        Highest priority (lowest level) first, then by resolution_id for stability.

        Args:
            sources: If provided, only return lineage paths that lead to these sources
            threshold: If provided, override this resolution's threshold

        Returns:
            List of tuples (resolution_id, source_config_id, threshold) ordered by
                priority.
        """
        with MBDB.get_session() as session:
            query = (
                select(
                    ResolutionFrom.parent,
                    SourceConfigs.source_config_id,
                    ResolutionFrom.truth_cache,
                )
                .join(
                    SourceConfigs,
                    ResolutionFrom.parent == SourceConfigs.resolution_id,
                    isouter=True,
                )
                .where(ResolutionFrom.child == self.resolution_id)
            )

            if sources:
                # Filter by source configs
                source_resolution_ids = [sc.resolution_id for sc in sources]

                descendant_ids = (
                    session.execute(
                        select(ResolutionFrom.child).where(
                            ResolutionFrom.parent.in_(source_resolution_ids)
                        )
                    )
                    .scalars()
                    .all()
                )

                query = query.where(
                    ResolutionFrom.parent.in_(source_resolution_ids + descendant_ids)
                )

            results = session.execute(
                query.order_by(ResolutionFrom.level.asc(), ResolutionFrom.parent.asc())
            ).all()

            # Get self's source config ID
            self_source_config_id = (
                self.source_config.source_config_id if self.source_config else None
            )

            # Threshold handling
            self_threshold = threshold if threshold is not None else self.truth

            # Add self at beginning (highest priority - level 0)
            return [(self.resolution_id, self_source_config_id, self_threshold)] + list(
                results
            )

    @classmethod
    def from_path(
        cls,
        path: ResolutionPath,
        res_type: ResolutionType | None = None,
        session: Session | None = None,
    ) -> "Resolutions":
        """Resolves a resolution name to a Resolution object.

        Args:
            path: The path of the resolution to resolve.
            res_type: A resolution type to use as filter.
            session: A session to get the resolution for updates.

        Raises:
            MatchboxResolutionNotFoundError: If the resolution doesn't exist.
        """
        query = (
            select(cls)
            .join(cls.run)
            .join(Runs.collection)
            .where(
                cls.name == path.name,
                Runs.run_id == path.run,
                Collections.name == path.collection,
            )
        )

        if res_type:
            query = query.where(cls.type == res_type.value)

        if session:
            resolution = session.execute(query).scalar()
        else:
            with MBDB.get_session() as session:
                resolution = session.execute(query).scalar()

        if resolution:
            return resolution

        raise MatchboxResolutionNotFoundError(
            message=f"No resolution {path} of type {res_type or 'any'}."
        )

    @classmethod
    def from_dto(
        cls, resolution: CommonResolution, path: ResolutionPath, session: Session
    ) -> "Resolutions":
        """Create a Resolutions instance from a Resolution DTO object.

        The resolution will be added to the session and flushed (but not committed).

        For model resolutions, lineage entries will be created automatically.

        Args:
            resolution: The Resolution DTO to convert
            path: The full resolution path
            session: Database session (caller must commit)

        Returns:
            A Resolutions ORM instance with ID and relationships established
        """
        # Find the run ID for the given collection and run ID
        run_obj = session.execute(
            select(Runs)
            .join(Collections)
            .where(Collections.name == path.collection, Runs.run_id == path.run)
        ).scalar_one_or_none()

        if not run_obj:
            raise MatchboxRunNotFoundError(number=path.run)

        # Check if resolution already exists within run
        existing_resolutions: Resolutions = run_obj.resolutions
        for res in existing_resolutions:
            if res.name == path.name:
                raise MatchboxResolutionAlreadyExists(
                    f"Resolution {path.name} already exists"
                )

        # Create new resolution
        resolution_orm = cls(
            run_id=run_obj.run_id,
            name=path.name,
            description=resolution.description,
            type=resolution.resolution_type.value,
            truth=resolution.truth,
            fingerprint=resolution.fingerprint,
        )
        session.add(resolution_orm)
        session.flush()  # Get resolution_id

        if resolution.resolution_type == ResolutionType.SOURCE:
            resolution_orm.source_config = SourceConfigs.from_dto(resolution.config)

        elif resolution.resolution_type == ResolutionType.MODEL:
            resolution_orm.model_config = ModelConfigs.from_dto(resolution.config)
            # Create lineage
            left_parent = cls.from_path(
                path=ResolutionPath(
                    collection=path.collection,
                    run=path.run,
                    name=resolution.config.left_query.point_of_truth,
                ),
                session=session,
            )
            cls._create_closure_entries(session, resolution_orm, left_parent)

            if resolution.config.type == ModelType.LINKER:
                right_parent = cls.from_path(
                    path=ResolutionPath(
                        collection=path.collection,
                        run=path.run,
                        name=resolution.config.right_query.point_of_truth,
                    ),
                    session=session,
                )
                cls._create_closure_entries(session, resolution_orm, right_parent)

        return resolution_orm

    def to_dto(self) -> CommonResolution:
        """Convert ORM resolution to a matchbox.common Resolution object."""
        if self.type == ResolutionType.SOURCE:
            config = self.source_config.to_dto()
        else:
            config = self.model_config.to_dto()

        return CommonResolution(
            description=self.description,
            truth=self.truth,
            resolution_type=ResolutionType(self.type),
            config=config,
            fingerprint=self.fingerprint,
        )

    @staticmethod
    def _create_closure_entries(
        session: Session, child: "Resolutions", parent: "Resolutions"
    ) -> None:
        """Create closure table entries for a parent-child relationship."""
        # Direct relationship
        session.add(
            ResolutionFrom(
                parent=parent.resolution_id,
                child=child.resolution_id,
                level=1,
                truth_cache=parent.truth,
            )
        )

        # Transitive closure
        ancestors = (
            session.query(ResolutionFrom)
            .filter(ResolutionFrom.child == parent.resolution_id)
            .all()
        )

        for ancestor in ancestors:
            session.add(
                ResolutionFrom(
                    parent=ancestor.parent,
                    child=child.resolution_id,
                    level=ancestor.level + 1,
                    truth_cache=ancestor.truth_cache,
                )
            )


class PKSpace(MBDB.MatchboxBase):
    """Table used to reserve ranges of primary keys."""

    __tablename__ = "pk_space"

    id = Column(BIGINT, primary_key=True)
    next_cluster_id = Column(BIGINT, nullable=False)
    next_cluster_keys_id = Column(BIGINT, nullable=False)

    @classmethod
    def initialise(cls) -> None:
        """Create PKSpace tracking row if not exists."""
        with MBDB.get_session() as session:
            init_statement = (
                insert(cls)
                .values(id=1, next_cluster_id=1, next_cluster_keys_id=1)
                .on_conflict_do_nothing()
            )

            session.execute(init_statement)
            session.commit()

    @classmethod
    def reserve_block(
        cls, table: Literal["clusters", "cluster_keys"], block_size: int
    ) -> int:
        """Atomically get next available ID for table, and increment it."""
        if block_size < 1:
            raise ValueError("Block size must be at least 1.")

        match table:
            case "clusters":
                next_id_col = "next_cluster_id"
            case "cluster_keys":
                next_id_col = "next_cluster_keys_id"

        with MBDB.get_session() as session:
            try:
                next_id = session.execute(
                    update(cls)
                    .values(**{next_id_col: getattr(cls, next_id_col) + block_size})
                    .returning(getattr(cls, next_id_col) - block_size)
                ).scalar_one()
                session.commit()
            except:
                session.rollback()
                raise

            return next_id


class SourceFields(CountMixin, MBDB.MatchboxBase):
    """Table for storing column details for SourceConfigs."""

    __tablename__ = "source_fields"

    # Columns
    field_id = Column(BIGINT, primary_key=True)
    source_config_id = Column(
        BIGINT,
        ForeignKey("source_configs.source_config_id", ondelete="CASCADE"),
        nullable=False,
    )
    index = Column(INTEGER, nullable=False)
    name = Column(TEXT, nullable=False)
    type = Column(TEXT, nullable=False)
    is_key = Column(BOOLEAN, nullable=False)

    # Relationships
    source_config = relationship(
        "SourceConfigs",
        back_populates="fields",
        foreign_keys=[source_config_id],
    )

    # Constraints and indices
    __table_args__ = (
        UniqueConstraint("source_config_id", "index", name="unique_index"),
        Index("ix_source_columns_source_config_id", "source_config_id"),
        Index(
            "ix_unique_key_field",
            "source_config_id",
            unique=True,
            postgresql_where=text("is_key = true"),
        ),
    )


class ClusterSourceKey(CountMixin, MBDB.MatchboxBase):
    """Table for storing source primary keys for clusters."""

    __tablename__ = "cluster_keys"

    # Columns
    key_id = Column(BIGINT, primary_key=True)
    cluster_id = Column(
        BIGINT, ForeignKey("clusters.cluster_id", ondelete="CASCADE"), nullable=False
    )
    source_config_id = Column(
        BIGINT,
        ForeignKey("source_configs.source_config_id", ondelete="CASCADE"),
        nullable=False,
    )
    key = Column(TEXT, nullable=False)

    # Relationships
    cluster = relationship("Clusters", back_populates="keys")
    source_config = relationship("SourceConfigs", back_populates="cluster_keys")

    # Constraints and indices
    __table_args__ = (
        Index("ix_cluster_keys_cluster_id", "cluster_id"),
        Index("ix_cluster_keys_keys", "key"),
        Index("ix_cluster_keys_source_config_id", "source_config_id"),
        UniqueConstraint("key_id", "source_config_id", name="unique_keys_source"),
    )


class SourceConfigs(CountMixin, MBDB.MatchboxBase):
    """Table of source_configs of data for Matchbox."""

    __tablename__ = "source_configs"

    # Columns
    source_config_id = Column(BIGINT, Identity(start=1), primary_key=True)
    resolution_id = Column(
        BIGINT,
        ForeignKey("resolutions.resolution_id", ondelete="CASCADE"),
        nullable=False,
    )
    location_type = Column(TEXT, nullable=False)
    location_name = Column(TEXT, nullable=False)
    extract_transform = Column(TEXT, nullable=False)

    @property
    def name(self) -> str:
        """Get the name of the related resolution."""
        return self.source_resolution.name

    # Relationships
    source_resolution = relationship("Resolutions", back_populates="source_config")
    fields = relationship(
        "SourceFields",
        back_populates="source_config",
        passive_deletes=True,
        cascade="all, delete-orphan",
    )
    key_field = relationship(
        "SourceFields",
        primaryjoin=(
            "and_(SourceConfigs.source_config_id == SourceFields.source_config_id, "
            "SourceFields.is_key == True)"
        ),
        viewonly=True,
        uselist=False,
    )
    index_fields = relationship(
        "SourceFields",
        primaryjoin=(
            "and_(SourceConfigs.source_config_id == SourceFields.source_config_id, "
            "SourceFields.is_key == False)"
        ),
        viewonly=True,
        order_by="SourceFields.index",
        collection_class=list,
    )
    cluster_keys = relationship(
        "ClusterSourceKey",
        back_populates="source_config",
        passive_deletes=True,
    )
    clusters = relationship(
        "Clusters",
        secondary=ClusterSourceKey.__table__,
        primaryjoin=(
            "SourceConfigs.source_config_id == ClusterSourceKey.source_config_id"
        ),
        secondaryjoin="ClusterSourceKey.cluster_id == Clusters.cluster_id",
        viewonly=True,
    )

    def __init__(
        self,
        key_field: SourceFields | None = None,
        index_fields: list[SourceFields] | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialise SourceConfigs with optional field objects."""
        super().__init__(**kwargs)

        # Add the key field and mark it as the key
        if key_field is not None:
            key_field.is_key = True
            key_field.index = 0
            self.fields.append(key_field)

        # Add index fields with proper indices
        if index_fields is not None:
            for idx, field in enumerate(index_fields):
                field.is_key = False
                field.index = idx + 1
                self.fields.append(field)

    @classmethod
    def list_all(cls) -> list["SourceConfigs"]:
        """Returns all source_configs in the database."""
        with MBDB.get_session() as session:
            return session.query(cls).all()

    @classmethod
    def from_dto(
        cls,
        config: CommonSourceConfig,
    ) -> "SourceConfigs":
        """Create a SourceConfigs instance from a Resolution DTO object."""
        # Create the SourceConfigs object
        return cls(
            location_type=str(config.location_config.type),
            location_name=str(config.location_config.name),
            extract_transform=config.extract_transform,
            key_field=SourceFields(
                index=0,
                name=config.key_field.name,
                type=config.key_field.type.value,
            ),
            index_fields=[
                SourceFields(
                    index=idx + 1,
                    name=field.name,
                    type=field.type.value,
                )
                for idx, field in enumerate(config.index_fields)
            ],
        )

    def to_dto(self) -> CommonSourceConfig:
        """Convert ORM source to a matchbox.common.SourceConfig object."""
        return CommonSourceConfig(
            location_config=LocationConfig(
                type=self.location_type, name=self.location_name
            ),
            extract_transform=self.extract_transform,
            key_field=CommonSourceField(
                name=self.key_field.name, type=self.key_field.type
            ),
            index_fields=[
                CommonSourceField(
                    name=field.name,
                    type=field.type,
                )
                for field in self.index_fields
            ],
        )


class ModelConfigs(CountMixin, MBDB.MatchboxBase):
    """Table of model configs for Matchbox."""

    __tablename__ = "model_configs"

    # Columns
    model_config_id = Column(BIGINT, Identity(start=1), primary_key=True)
    resolution_id = Column(
        BIGINT,
        ForeignKey("resolutions.resolution_id", ondelete="CASCADE"),
        nullable=False,
    )
    model_class = Column(TEXT, nullable=False)
    model_settings = Column(JSONB, nullable=False)
    left_query = Column(JSONB, nullable=False)
    right_query = Column(JSONB, nullable=True)

    @property
    def name(self) -> str:
        """Get the name of the related resolution."""
        return self.model_resolution.name

    # Relationships
    model_resolution = relationship("Resolutions", back_populates="model_config")

    @classmethod
    def list_all(cls) -> list["SourceConfigs"]:
        """Returns all model_configs in the database."""
        with MBDB.get_session() as session:
            return session.query(cls).all()

    @classmethod
    def from_dto(
        cls,
        config: CommonModelConfig,
    ) -> "ModelConfigs":
        """Create a SourceConfigs instance from a Resolution DTO object."""
        # Create the SourceConfigs object
        return cls(
            model_class=config.model_class,
            model_settings=config.model_settings,
            left_query=config.left_query.model_dump_json(),
            right_query=(
                None if not config.right_query else config.right_query.model_dump_json()
            ),
        )

    def to_dto(self) -> CommonModelConfig:
        """Convert ORM source to a matchbox.common.ModelConfig object."""
        return CommonModelConfig(
            type=ModelType.LINKER if self.right_query else ModelType.DEDUPER,
            model_class=self.model_class,
            model_settings=self.model_settings,
            left_query=json.loads(self.left_query),
            right_query=json.loads(self.right_query) if self.right_query else None,
        )


class Contains(CountMixin, MBDB.MatchboxBase):
    """Cluster lineage table."""

    __tablename__ = "contains"

    # Columns
    root = Column(
        BIGINT, ForeignKey("clusters.cluster_id", ondelete="CASCADE"), primary_key=True
    )
    leaf = Column(
        BIGINT, ForeignKey("clusters.cluster_id", ondelete="CASCADE"), primary_key=True
    )

    # Constraints and indices
    __table_args__ = (
        CheckConstraint("root != leaf", name="no_self_containment"),
        Index("ix_contains_root_leaf", "root", "leaf"),
        Index("ix_contains_leaf_root", "leaf", "root"),
    )


class Clusters(CountMixin, MBDB.MatchboxBase):
    """Table of indexed data and clusters that match it."""

    __tablename__ = "clusters"

    # Columns
    cluster_id = Column(BIGINT, primary_key=True)
    cluster_hash = Column(BYTEA, nullable=False)

    # Relationships
    keys = relationship(
        "ClusterSourceKey",
        back_populates="cluster",
        passive_deletes=True,
    )
    probabilities = relationship(
        "Probabilities",
        back_populates="proposes",
        passive_deletes=True,
    )
    leaves = relationship(
        "Clusters",
        secondary=Contains.__table__,
        primaryjoin="Clusters.cluster_id == Contains.root",
        secondaryjoin="Clusters.cluster_id == Contains.leaf",
        backref="roots",
    )
    # Add relationship to SourceConfigs through ClusterSourceKey
    source_configs = relationship(
        "SourceConfigs",
        secondary=ClusterSourceKey.__table__,
        primaryjoin="Clusters.cluster_id == ClusterSourceKey.cluster_id",
        secondaryjoin=(
            "ClusterSourceKey.source_config_id == SourceConfigs.source_config_id"
        ),
        viewonly=True,
    )

    # Constraints and indices
    __table_args__ = (UniqueConstraint("cluster_hash", name="clusters_hash_key"),)


class Users(CountMixin, MBDB.MatchboxBase):
    """Table of identities of human validators."""

    __tablename__ = "users"

    # Columns
    user_id = Column(BIGINT, primary_key=True)
    name = Column(TEXT, nullable=False)

    judgements = relationship("EvalJudgements", back_populates="user")

    __table_args__ = (UniqueConstraint("name", name="user_name_unique"),)


class EvalJudgements(CountMixin, MBDB.MatchboxBase):
    """Table of evaluation judgements produced by human validators."""

    __tablename__ = "eval_judgements"

    # Columns
    judgement_id = Column(BIGINT, primary_key=True)
    user_id = Column(
        BIGINT, ForeignKey("users.user_id", ondelete="CASCADE"), nullable=False
    )
    endorsed_cluster_id = Column(
        BIGINT, ForeignKey("clusters.cluster_id", ondelete="CASCADE"), nullable=False
    )
    shown_cluster_id = Column(
        BIGINT, ForeignKey("clusters.cluster_id", ondelete="CASCADE"), nullable=False
    )
    timestamp = Column(DateTime(timezone=True), nullable=False)

    # Relationships
    user = relationship("Users", back_populates="judgements")


class Probabilities(CountMixin, MBDB.MatchboxBase):
    """Table of probabilities that a cluster is correct, according to a resolution."""

    __tablename__ = "probabilities"

    # Columns
    resolution_id = Column(
        BIGINT,
        ForeignKey("resolutions.resolution_id", ondelete="CASCADE"),
        primary_key=True,
    )
    cluster_id = Column(
        BIGINT, ForeignKey("clusters.cluster_id", ondelete="CASCADE"), primary_key=True
    )
    probability = Column(SMALLINT, nullable=False)

    # Relationships
    proposed_by = relationship("Resolutions", back_populates="probabilities")
    proposes = relationship("Clusters", back_populates="probabilities")

    # Constraints
    __table_args__ = (
        CheckConstraint("probability BETWEEN 0 AND 100", name="valid_probability"),
        Index("ix_probabilities_resolution", "resolution_id"),
    )


class Results(CountMixin, MBDB.MatchboxBase):
    """Table of results for a resolution.

    Stores the raw left/right probabilities created by a model.
    """

    __tablename__ = "results"

    # Columns
    result_id = Column(BIGINT, primary_key=True, autoincrement=True)
    resolution_id = Column(
        BIGINT,
        ForeignKey("resolutions.resolution_id", ondelete="CASCADE"),
        nullable=False,
    )
    left_id = Column(
        BIGINT, ForeignKey("clusters.cluster_id", ondelete="CASCADE"), nullable=False
    )
    right_id = Column(
        BIGINT, ForeignKey("clusters.cluster_id", ondelete="CASCADE"), nullable=False
    )
    probability = Column(SMALLINT, nullable=False)

    # Relationships
    proposed_by = relationship("Resolutions", back_populates="results")

    # Constraints
    __table_args__ = (
        Index("ix_results_resolution", "resolution_id"),
        CheckConstraint("probability BETWEEN 0 AND 100", name="valid_probability"),
        UniqueConstraint("resolution_id", "left_id", "right_id"),
    )



"""Add migrations for first eval tables.

Revision ID: dd0c3a9ecdf9
Revises: 3754ae042254
Create Date: 2025-05-14 11:33:54.343350

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "dd0c3a9ecdf9"
down_revision: str | None = "3754ae042254"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    op.create_table(
        "users",
        sa.Column("user_id", sa.BIGINT(), nullable=False),
        sa.Column("name", sa.TEXT(), nullable=False),
        sa.PrimaryKeyConstraint("user_id"),
        sa.UniqueConstraint("name", name="user_name_unique"),
        schema="mb",
    )
    op.create_table(
        "eval_judgements",
        sa.Column("judgement_id", sa.BIGINT(), nullable=False),
        sa.Column("user_id", sa.BIGINT(), nullable=False),
        sa.Column("endorsed_cluster_id", sa.BIGINT(), nullable=False),
        sa.Column("shown_cluster_id", sa.BIGINT(), nullable=False),
        sa.Column("timestamp", sa.DateTime(timezone=True), nullable=False),
        sa.ForeignKeyConstraint(
            ["endorsed_cluster_id"], ["mb.clusters.cluster_id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["shown_cluster_id"], ["mb.clusters.cluster_id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(["user_id"], ["mb.users.user_id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("judgement_id"),
        schema="mb",
    )


def downgrade() -> None:
    """Downgrade schema."""
    op.drop_table("eval_judgements", schema="mb")
    op.drop_table("users", schema="mb")



import pytest

from matchbox.common.dtos import (
    Collection,
    CollectionName,
    Match,
    ModelResolutionName,
    ResolutionName,
    ResolutionPath,
    SourceResolutionName,
)
from matchbox.common.exceptions import MatchboxNameError


def test_validate_names() -> None:
    """Names are validated when they're instantiated."""
    name_classes = [
        CollectionName,
        ModelResolutionName,
        ResolutionName,
        SourceResolutionName,
    ]

    [NameClass("Valid.name_-") for NameClass in name_classes]

    for NameClass in name_classes:
        with pytest.raises(MatchboxNameError):
            NameClass("invalid name")


def test_match_validates() -> None:
    """Match objects are validated when they're instantiated."""
    source_path = ResolutionPath(name="source", collection="default", run=1)
    target_path = ResolutionPath(name="target", collection="default", run=1)
    Match(
        cluster=1,
        source=source_path,
        source_id={"a"},
        target=target_path,
        target_id={"b"},
    )

    # Missing source_id with target_id
    with pytest.raises(ValueError):
        Match(cluster=1, source=source_path, target=target_path, target_id={"b"})

    # Missing cluster with target_id
    with pytest.raises(ValueError):
        Match(source=source_path, source_id={"a"}, target=target_path, target_id={"b"})

    # Missing source_id with cluster
    with pytest.raises(ValueError):
        Match(cluster=1, source=source_path, target=target_path)


def test_validate_collection() -> None:
    """Default run name needs to be within all runs."""
    # No default
    Collection(runs=[1, 2])

    # Valid default
    Collection(default_run=2, runs=[1, 2])

    with pytest.raises(ValueError):
        Collection(default_run=2, runs=[1])



"""CLI commands for entity evaluation."""

import logging
from typing import Annotated

import typer
from sqlalchemy import create_engine

from matchbox.client.cli.eval.app import EntityResolutionApp
from matchbox.client.dags import DAG


def evaluate(
    collection: Annotated[
        str, typer.Option("--collection", "-c", help="Collection name (required)")
    ],
    resolution: Annotated[
        str | None,
        typer.Option(
            "--resolution",
            "-r",
            help="Resolution name (defaults to collection's final_step)",
        ),
    ] = None,
    pending: Annotated[
        bool,
        typer.Option(
            "--pending",
            "-p",
            help="Whether to evaluate the pending DAG, instead of the default",
        ),
    ] = False,
    user: Annotated[
        str | None,
        typer.Option(
            "--user", "-u", help="Username for authentication (overrides settings)"
        ),
    ] = None,
    warehouse: Annotated[
        str | None,
        typer.Option(
            "--warehouse",
            "-w",
            help="Warehouse database connection string (e.g. postgresql://user:pass@host/db)",
        ),
    ] = None,
    log_file: Annotated[
        str | None,
        typer.Option(
            "--log",
            help="Log file path to redirect all logging output (keeps UI clean)",
        ),
    ] = None,
) -> None:
    """Start the interactive entity resolution evaluation tool.

    Requires a warehouse connection to fetch source data for evaluation clusters.

    Example:
        matchbox eval --collection companies --warehouse postgresql://user:pass@localhost/warehouse
    """
    # Set up logging redirect if --log specified
    if log_file:
        _setup_logging_redirect(log_file)

    # Create warehouse engine and location (required for loading DAG)
    if not warehouse:
        raise typer.BadParameter(
            "Warehouse connection string is required. "
            "Provide via --warehouse or -w flag."
        )

    # Create engine from connection string (with password)
    warehouse_engine = create_engine(warehouse)

    # Load DAG from server
    if pending:
        dag: DAG = DAG(name=collection).load_pending()
    else:
        dag: DAG = DAG(name=collection).load_default()

    # Attach warehouse to all objects
    dag: DAG = dag.set_client(warehouse_engine)

    # Get resolution name from --resolution or DAG's final_step
    model = dag.get_model(resolution) if resolution is not None else dag.final_step

    try:
        # Create app with loaded DAG (not warehouse string)
        app = EntityResolutionApp(
            resolution=model.resolution_path.name,
            user=user,
            dag=dag,
            show_help=True,
        )
        app.run()
    except KeyboardInterrupt:
        # Handle Ctrl+C gracefully - exit with success code
        pass


def _setup_logging_redirect(log_file_path: str) -> None:
    """Redirect all logging output to file to keep Textual UI clean."""
    # Get the root logger to capture ALL logging (including third-party)
    root_logger = logging.getLogger()

    # Remove any existing handlers (console handlers that print to screen)
    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)

    # Set up file handler for all logs
    file_handler = logging.FileHandler(log_file_path, mode="w")

    # Use detailed format since these are going to file, not screen
    formatter = logging.Formatter("%(asctime)s [%(levelname)s] %(name)s: %(message)s")
    file_handler.setFormatter(formatter)

    # Add file handler to root logger
    root_logger.addHandler(file_handler)
    root_logger.setLevel(logging.INFO)  # Capture INFO and above



"""Integration tests with real scenario data - comprehensive behaviour testing."""

from collections.abc import Callable
from functools import partial
from unittest.mock import Mock

import pytest
from sqlalchemy import Engine

from matchbox.client.cli.eval.app import EntityResolutionApp, EvaluationQueue
from matchbox.client.dags import DAG
from matchbox.common.dtos import ModelResolutionPath
from matchbox.common.factories.scenarios import setup_scenario
from matchbox.server.base import MatchboxDBAdapter

backends = [
    pytest.param("matchbox_postgres", id="postgres"),
]


@pytest.fixture(scope="function")
def backend_instance(request: pytest.FixtureRequest, backend: str) -> MatchboxDBAdapter:
    """Create a fresh backend instance for each test."""
    backend_obj = request.getfixturevalue(backend)
    backend_obj.clear(certain=True)
    return backend_obj


class TestEvaluationQueue:
    """Unit tests for EvaluationQueue - no app or mocking needed."""

    def test_queue_initialises_empty(self) -> None:
        """Test queue starts empty."""
        queue = EvaluationQueue()

        assert queue.total_count == 0
        assert queue.current is None

    def test_add_items_increases_count(self) -> None:
        """Test adding items increases count."""
        queue = EvaluationQueue()
        items = [Mock(cluster_id=1), Mock(cluster_id=2)]

        added = queue.add_items(items)

        assert added == 2
        assert queue.total_count == 2
        assert queue.current is items[0]

    def test_add_items_prevents_duplicates(self) -> None:
        """Test that duplicate cluster IDs are not added."""
        queue = EvaluationQueue()
        item1 = Mock(cluster_id=1)
        item2 = Mock(cluster_id=1)  # Duplicate

        queue.add_items([item1])
        added = queue.add_items([item2])

        assert added == 0
        assert queue.total_count == 1

    def test_add_items_handles_empty_list(self) -> None:
        """Test that adding empty list returns 0."""
        queue = EvaluationQueue()

        added = queue.add_items([])

        assert added == 0
        assert queue.total_count == 0

    def test_skip_rotates_deque(self) -> None:
        """Test that skip moves current to back."""
        queue = EvaluationQueue()
        item1 = Mock(cluster_id=1)
        item2 = Mock(cluster_id=2)
        queue.items.extend([item1, item2])

        queue.skip_current()

        assert queue.current is item2
        assert queue.items[1] is item1

    def test_skip_with_single_item_does_nothing(self) -> None:
        """Test that skip with one item doesn't rotate."""
        queue = EvaluationQueue()
        item1 = Mock(cluster_id=1)
        queue.items.append(item1)

        queue.skip_current()

        assert queue.current is item1
        assert queue.total_count == 1

    def test_remove_current_pops_front(self) -> None:
        """Test that remove_current removes from front."""
        queue = EvaluationQueue()
        item1 = Mock(cluster_id=1)
        item2 = Mock(cluster_id=2)
        queue.items.extend([item1, item2])

        removed = queue.remove_current()

        assert removed is item1
        assert queue.total_count == 1
        assert queue.current is item2

    def test_remove_current_on_empty_returns_none(self) -> None:
        """Test that remove_current on empty queue returns None."""
        queue = EvaluationQueue()

        removed = queue.remove_current()

        assert removed is None
        assert queue.total_count == 0


@pytest.mark.parametrize("backend", backends)
@pytest.mark.docker
class TestScenarioIntegration:
    """Integration tests using real scenario data with backend."""

    @pytest.fixture
    def test_resolution(self) -> ModelResolutionPath:
        """Create test resolution path."""
        return ModelResolutionPath(
            collection="test_collection", run=1, name="test_resolution"
        )

    @pytest.fixture(autouse=True)
    def setup(self, backend_instance: str, sqlite_warehouse: Engine) -> None:
        """Set up test fixtures."""
        self.backend: MatchboxDBAdapter = backend_instance
        self.warehouse_engine: Engine = sqlite_warehouse
        self.scenario: Callable = partial(setup_scenario, warehouse=sqlite_warehouse)

    @pytest.mark.asyncio
    async def test_complete_evaluation_workflow(self) -> None:
        """Test complete evaluation workflow: start, load, label, clear, skip, submit.

        This comprehensive test covers the main user journey through the evaluation app:

        - App initialisation and sample loading
        - Keyboard-driven assignment workflow (letter → digit)
        - Clearing assignments
        - Skipping items in the queue
        - Submitting incomplete assignments (warning)
        - Completing and submitting assignments (saves judgement)
        - Status updates and help modal

        Additional edge cases (e.g., clusters too large for screen) should be
        separate tests as they may require different setup or scenarios.
        """
        with self.scenario(self.backend, "dedupe") as dag:
            model_name: str = "naive_test_crn"

            loaded_dag: DAG = (
                DAG(str(dag.dag.name)).load_pending().set_client(self.warehouse_engine)
            )

            app = EntityResolutionApp(
                resolution=model_name,
                num_samples=3,
                user="test_user",
                dag=loaded_dag,
            )

            async with app.run_test() as pilot:
                await pilot.pause()

                # 1. Verify app initialisation and sample loading
                assert app.is_running
                assert app.user_name == "test_user"
                assert app.user_id is not None
                assert pilot.app.query("Footer")
                assert app.queue.total_count > 0
                assert app.queue.current is not None
                assert app.queue.current.cluster_id is not None

                # 2. Test keyboard workflow: letter then digit
                await pilot.press("a")
                await pilot.pause()
                assert app.current_group == "a"

                await pilot.press("1")
                await pilot.pause()
                current = app.queue.current
                assert current is not None
                assert 0 in current.assignments
                assert current.assignments[0] == "a"

                # 3. Test making additional assignments
                await pilot.press("b")
                await pilot.press("2")
                await pilot.pause()
                assert len(current.assignments) > 1

                # 4. Test status updates reactively
                assert isinstance(app.status, tuple)
                assert len(app.status) == 2

                # 5. Test clearing assignments
                await pilot.press("escape")
                await pilot.pause()
                assert len(current.assignments) == 0
                assert app.current_group == ""

                # 6. Test skip workflow
                first_item = app.queue.current
                await pilot.press("shift+right")
                await pilot.pause()
                assert app.queue.current is not first_item
                assert app.queue.items[-1] is first_item

                # 7. Test submitting incomplete assignment shows warning
                initial_count = app.queue.total_count
                await pilot.press("space")
                await pilot.pause()
                assert app.queue.total_count == initial_count
                status_message = app.status[0].lower()
                assert "incomplete" in status_message

                # 8. Test completing and submitting assignment sends judgement
                current = app.queue.current
                assert current is not None
                num_columns = len(current.display_columns)
                for i in range(num_columns):
                    await pilot.press("a")
                    await pilot.press(str((i % 9) + 1))
                    await pilot.pause()

                initial_judgements, _ = self.backend.get_judgements()
                initial_judgement_count = len(initial_judgements)

                await pilot.press("space")
                await pilot.pause()

                final_judgements, _ = self.backend.get_judgements()
                final_judgement_count = len(final_judgements)
                assert final_judgement_count == initial_judgement_count + 1

                # 9. Test help modal
                initial_stack_size = len(pilot.app.screen_stack)
                await pilot.press("f1")
                await pilot.pause()
                assert len(pilot.app.screen_stack) > initial_stack_size

    @pytest.mark.asyncio
    async def test_navigation(self) -> None:
        """Test page and tab navigation."""
        with self.scenario(self.backend, "mega") as dag:
            model_name: str = "mega_product_linker"

            loaded_dag: DAG = (
                DAG(str(dag.dag.name)).load_pending().set_client(self.warehouse_engine)
            )

            app = EntityResolutionApp(
                resolution=model_name,
                num_samples=20,
                user="test_user",
                dag=loaded_dag,
            )

            async with app.run_test() as pilot:
                await pilot.resize_terminal(120, 40)
                await pilot.pause()

                initial_page = app.current_page_by_tab.get(0, 0)

                await pilot.press("down")
                await pilot.pause()
                assert app.current_page_by_tab.get(0, 0) == initial_page + 1

                await pilot.press("up")
                await pilot.pause()
                assert app.current_page_by_tab.get(0, 0) == initial_page

                initial_tab = app._get_current_tab_index()

                await pilot.press("right")
                await pilot.pause()
                assert app._get_current_tab_index() == initial_tab + 1

                await pilot.press("left")
                await pilot.pause()
                assert app._get_current_tab_index() == initial_tab



"""Common database utilities for Matchbox."""

from collections.abc import Callable, Iterator
from enum import StrEnum
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    TypeAlias,
    TypeVar,
    overload,
)

import polars as pl
from pandas import DataFrame as PandasDataFrame
from polars import DataFrame as PolarsDataFrame
from pyarrow import Table as ArrowTable
from sqlalchemy.engine import Engine

if TYPE_CHECKING:
    from adbc_driver_postgresql.dbapi import Connection as ADBCConnection
else:
    ADBCConnection = Any


class QueryReturnType(StrEnum):
    """Enumeration of dataframe types to return from query."""

    PANDAS = "pandas"
    POLARS = "polars"
    ARROW = "arrow"


QueryReturnClass: TypeAlias = ArrowTable | PandasDataFrame | PolarsDataFrame

T = TypeVar("T")


@overload
def sql_to_df(
    stmt: str,
    connection: Engine | ADBCConnection,
    return_type: QueryReturnType,
    *,
    return_batches: Literal[False] = False,
    batch_size: int | None = None,
    rename: dict[str, str] | Callable | None = None,
    schema_overrides: dict[str, pl.DataType] | None = None,
    execute_options: dict[str, Any] | None = None,
) -> QueryReturnClass: ...


@overload
def sql_to_df(
    stmt: str,
    connection: Engine | ADBCConnection,
    return_type: QueryReturnType,
    *,
    return_batches: Literal[True],
    batch_size: int | None = None,
    rename: dict[str, str] | Callable | None = None,
    schema_overrides: dict[str, pl.DataType] | None = None,
    execute_options: dict[str, Any] | None = None,
) -> Iterator[QueryReturnClass]: ...


def sql_to_df(
    stmt: str,
    connection: Engine | ADBCConnection,
    return_type: QueryReturnType = QueryReturnType.PANDAS,
    *,
    return_batches: bool = False,
    batch_size: int | None = None,
    rename: dict[str, str] | Callable | None = None,
    schema_overrides: dict[str, pl.DataType] | None = None,
    execute_options: dict[str, Any] | None = None,
) -> QueryReturnClass | Iterator[QueryReturnClass]:
    """Executes the given SQLAlchemy statement or SQL string using Polars.

    Args:
        stmt: A SQL string to be executed.
        connection: A SQLAlchemy Engine object or
            ADBC connection.
        return_type: The type of the return value. One of "arrow", "pandas",
            or "polars".
        return_batches: If True, return an iterator that yields each batch
            separately. If False, return a single DataFrame with all results.
            Default is False.
        batch_size: Indicate the size of each batch when processing
            data in batches. Default is None.
        rename: A dictionary mapping old column
            names to new column names, or a callable that takes a DataFrame and
            returns a DataFrame with renamed columns. Default is None.
        schema_overrides: A dictionary mapping column names
            to dtypes. Default is None.
        execute_options: These options will be passed through
            into the underlying query execution method as kwargs. Default is None.

    Returns:
        If return_batches is False: A dataframe of the query results in the specified
            format.
        If return_batches is True: An iterator of dataframes in the specified format.

    Raises:
        ValueError:

            * If the connection is not properly configured or if an unsupported
                return type is specified.
            * If batch_size and return_batches are either both set or both unset.

    """
    if not batch_size and return_batches:
        raise ValueError("A batch size must be specified if return_batches is True")

    if batch_size and not return_batches:
        raise ValueError("Cannot set a batch size if return_batches if False")

    def _to_format(results: PolarsDataFrame) -> QueryReturnClass:
        """Convert the results to the specified format."""
        if rename:
            results = results.rename(rename)

        match return_type:
            case QueryReturnType.POLARS:
                return results
            case QueryReturnType.PANDAS:
                return results.to_pandas()
            case QueryReturnType.ARROW:
                return results.to_arrow()
            case _:
                raise ValueError("Unknown return type specified.")

    res = pl.read_database(
        query=stmt,
        connection=connection,
        iter_batches=bool(batch_size),
        batch_size=batch_size,
        schema_overrides=schema_overrides,
        execute_options=execute_options,
    )

    if return_batches:
        return (_to_format(batch) for batch in res)

    return _to_format(res)



import polars as pl
import pytest
from polars.testing import assert_frame_equal
from sqlalchemy import Engine
from sqlalchemy.exc import OperationalError

from matchbox.client.locations import RelationalDBLocation
from matchbox.common.dtos import (
    DataTypes,
    LocationType,
)
from matchbox.common.exceptions import MatchboxSourceExtractTransformError
from matchbox.common.factories.sources import (
    FeatureConfig,
    source_factory,
    source_from_tuple,
)


def test_relational_db_location_instantiation(
    sqlite_in_memory_warehouse: Engine,
) -> None:
    """Test that RelationalDBLocation can be instantiated with valid parameters."""
    location = RelationalDBLocation(name="dbname")
    assert location.config.type == LocationType.RDBMS
    assert location.config.name == "dbname"

    # Client can be set and validated
    assert location.client is None

    with pytest.raises(ValueError):
        location.set_client(12)

    assert (
        location.set_client(sqlite_in_memory_warehouse).client
        == sqlite_in_memory_warehouse
    )


@pytest.mark.parametrize(
    ["sql", "dialects"],
    [
        pytest.param("SELECT * FROM test_table", "all", id="valid-select"),
        pytest.param(
            "SELECT id, name FROM test_table WHERE id > 1", "all", id="valid-where"
        ),
        pytest.param("SLECT * FROM test_table", "none", id="invalid-syntax"),
        pytest.param("", "none", id="empty-string"),
        pytest.param("ALTER TABLE test_table", "none", id="alter-sql"),
        pytest.param(
            "INSERT INTO users (name, age) VALUES ('John', '25')",
            "none",
            id="insert-sql",
        ),
        pytest.param("DROP TABLE test_table", "none", id="drop-sql"),
        pytest.param("SELECT * FROM users /* with a comment */", "all", id="comment"),
        pytest.param(
            "WITH user_cte AS (SELECT * FROM users) SELECT * FROM user_cte",
            "all",
            id="valid-with",
        ),
        pytest.param(
            (
                "WITH user_cte AS (SELECT * FROM users) "
                "INSERT INTO temp_users SELECT * FROM user_cte"
            ),
            "none",
            id="invalid-with",
        ),
        pytest.param(
            "SELECT * FROM users; DROP TABLE users;", "none", id="multiple-statements"
        ),
        pytest.param("SELECT * INTO new_table FROM users", "none", id="select-into"),
        pytest.param(
            """
            WITH updated_rows AS (
                UPDATE employees
                SET salary = salary * 1.1
                WHERE department = 'Sales'
                RETURNING *
            )
            SELECT * FROM updated_rows;
            """,
            "none",
            id="non-query-cte",
        ),
        pytest.param(
            """
            SELECT foo, bar FROM baz
            UNION
            SELECT foo, bar FROM qux;
            """,
            "all",
            id="valid-union",
        ),
        pytest.param(
            """
            SELECT 'ciao' ~ 'hello'
            """,
            "postgres",
            id="postgres-only",
        ),
        pytest.param("""select `name` from user""", "sqlite", id="sqlite-only"),
    ],
)
def test_relational_db_extract_transform(
    sql: str, dialects: str, postgres_warehouse: Engine, sqlite_warehouse: Engine
) -> None:
    """Test SQL validation in validate_extract_transform."""

    if dialects == "none":
        invalid_clients = [postgres_warehouse, sqlite_warehouse]
        valid_clients = []
    if dialects == "all":
        invalid_clients = []
        valid_clients = [postgres_warehouse, sqlite_warehouse]
    if dialects == "postgres":
        invalid_clients = [sqlite_warehouse]
        valid_clients = [postgres_warehouse]
    if dialects == "sqlite":
        invalid_clients = [postgres_warehouse]
        valid_clients = [sqlite_warehouse]

    # Dialect-agnostic check
    if dialects == "none":
        with pytest.raises(MatchboxSourceExtractTransformError):
            RelationalDBLocation(name="dbname").validate_extract_transform(sql)
    elif dialects == "all":
        RelationalDBLocation(name="dbname").validate_extract_transform(sql)

    # Dialect-specific checks
    for client in valid_clients:
        RelationalDBLocation(name="dbname").set_client(
            client
        ).validate_extract_transform(sql)

    for client in invalid_clients:
        with pytest.raises(MatchboxSourceExtractTransformError):
            RelationalDBLocation(name="dbname").set_client(
                client
            ).validate_extract_transform(sql)


def test_relational_db_infer_types(sqlite_warehouse: Engine) -> None:
    """Test that types are inferred correctly from the extract transform SQL."""
    source_testkit = source_from_tuple(
        data_tuple=(
            {"foo": "10", "bar": None},
            {"foo": "foo_val", "bar": None},
            {"foo": None, "bar": 10},
        ),
        data_keys=["a", "b", "c"],
        name="source",
        engine=sqlite_warehouse,
    ).write_to_location()
    location = RelationalDBLocation(name="dbname").set_client(sqlite_warehouse)

    query = f"""
        select key as renamed_key, foo, bar from
        (select key, foo, bar from {source_testkit.name});
    """

    inferred_types = location.infer_types(query)

    assert len(inferred_types) == 3
    assert inferred_types["renamed_key"] == DataTypes.STRING
    assert inferred_types["foo"] == DataTypes.STRING
    assert inferred_types["bar"] == DataTypes.INT64


def test_relational_db_execute(sqlite_warehouse: Engine) -> None:
    """Test executing a query and returning results using a real SQLite database."""
    features = [
        FeatureConfig(name="company", base_generator="company"),
        FeatureConfig(name="employees", base_generator="random_int"),
    ]

    source_testkit = source_factory(
        features=features, n_true_entities=10, engine=sqlite_warehouse
    ).write_to_location()
    location = RelationalDBLocation(name="dbname").set_client(sqlite_warehouse)

    sql = f"select key, upper(company) up_company, employees from {source_testkit.name}"

    batch_size = 2

    # Execute the query
    results = list(location.execute(sql, batch_size))

    # Right number of batches and total rows
    assert len(results[0]) == batch_size
    combined_df: pl.DataFrame = pl.concat(results)
    assert len(combined_df) == 10

    # Right fields, types and transformations
    assert set(combined_df.columns) == {"key", "up_company", "employees"}
    assert combined_df["employees"].dtype == pl.Int64
    sample_str = combined_df.select("up_company").row(0)[0]
    assert sample_str == sample_str.upper()

    # Try overriding schema
    overridden_results = list(
        location.execute(sql, batch_size, schema_overrides={"employees": pl.String})
    )
    assert overridden_results[0]["employees"].dtype == pl.String

    # Try query with filter
    keys_to_filter = source_testkit.data["key"][:2].to_pylist()
    filtered_results = pl.concat(
        location.execute(sql, batch_size, keys=("key", keys_to_filter))
    )
    assert len(filtered_results) == 2

    # Filtering by no keys has no effect
    unfiltered_results = pl.concat(location.execute(sql, batch_size, keys=("key", [])))
    assert_frame_equal(unfiltered_results, combined_df)


def test_relational_db_execute_invalid(sqlite_warehouse: Engine) -> None:
    """Test that invalid queries are handled correctly when executing."""
    location = RelationalDBLocation(name="dbname").set_client(sqlite_warehouse)

    # Invalid SQL query
    sql = "SELECT * FROM nonexistent_table"

    # Should raise an exception when executed
    with pytest.raises(OperationalError):
        list(location.execute(sql, batch_size=10))



"""Add ModelConfigs.

Revision ID: c4cb937d00f4
Revises: b38d61ab11cc
Create Date: 2025-09-11 12:30:44.246293

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = "c4cb937d00f4"
down_revision: str | None = "b38d61ab11cc"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    op.create_table(
        "model_configs",
        sa.Column(
            "model_config_id",
            sa.BIGINT(),
            sa.Identity(always=False, start=1),
            nullable=False,
        ),
        sa.Column("resolution_id", sa.BIGINT(), nullable=False),
        sa.Column("model_class", sa.TEXT(), nullable=False),
        sa.Column(
            "model_settings", postgresql.JSONB(astext_type=sa.Text()), nullable=False
        ),
        sa.Column(
            "left_query", postgresql.JSONB(astext_type=sa.Text()), nullable=False
        ),
        sa.Column(
            "right_query", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.ForeignKeyConstraint(
            ["resolution_id"], ["mb.resolutions.resolution_id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("model_config_id"),
        schema="mb",
    )


def downgrade() -> None:
    """Downgrade schema."""
    op.drop_table("model_configs", schema="mb")



"""Switch from location URI to name.

Revision ID: 7a2d1b10ac0f
Revises: dd0c3a9ecdf9
Create Date: 2025-07-29 07:52:36.312694

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "7a2d1b10ac0f"
down_revision: str | None = "dd0c3a9ecdf9"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    # Add default value for rows without location names
    op.add_column(
        "source_configs",
        sa.Column(
            "location_name",
            sa.TEXT(),
            nullable=False,
            server_default="legacy",
        ),
        schema="mb",
    )
    # Remove default value for location names
    op.alter_column("source_configs", "location_name", server_default=None, schema="mb")
    op.drop_column("source_configs", "location_uri", schema="mb")


def downgrade() -> None:
    """Downgrade schema."""
    # Add default value for rows without location URIs
    op.add_column(
        "source_configs",
        sa.Column(
            "location_uri",
            sa.TEXT(),
            autoincrement=False,
            nullable=False,
            server_default="sqlite:///:memory:",
        ),
        schema="mb",
    )
    # Remove default value for location URIs
    op.alter_column("source_configs", "location_uri", server_default=None, schema="mb")
    op.drop_column("source_configs", "location_name", schema="mb")



"""Objects representing the results of running a model client-side."""

from collections.abc import Hashable
from typing import ParamSpec, TypeVar

import polars as pl
from pydantic import ConfigDict

from matchbox.common.arrow import SCHEMA_RESULTS
from matchbox.common.hash import IntMap
from matchbox.common.transform import to_clusters

T = TypeVar("T", bound=Hashable)
P = ParamSpec("P")
R = TypeVar("R")


class Results:
    """Results of a model run.

    Contains:

    * The probabilities of each pair being a match
    * (Optional) The clusters of connected components at each threshold
    * (Optional) The leaf_id mapping to trace results back to source clusters

    Allows users to easily interrogate the outputs of models, explore decisions on
    choosing thresholds for clustering, and upload the results to Matchbox.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    probabilities: pl.DataFrame
    _clusters: pl.DataFrame | None = None

    def __init__(
        self,
        probabilities: pl.DataFrame,
        left_root_leaf: pl.DataFrame | None = None,
        right_root_leaf: pl.DataFrame | None = None,
    ) -> None:
        """Initialises and validates results."""
        self.left_root_leaf = None
        self.right_root_leaf = None

        if left_root_leaf is not None:
            self.left_root_leaf = left_root_leaf
        if right_root_leaf is not None:
            self.right_root_leaf = right_root_leaf

        if not isinstance(probabilities, pl.DataFrame):
            raise ValueError(f"Expected a polars DataFrame, got {type(probabilities)}.")

        expected_fields = set(SCHEMA_RESULTS.names)
        if set(probabilities.columns) != expected_fields:
            raise ValueError(
                f"Expected {expected_fields}.\nFound {set(probabilities.column_names)}."
            )

        # Handle empty tables
        if probabilities.height == 0:
            probabilities = pl.DataFrame(schema=pl.Schema(SCHEMA_RESULTS))

        # Process probability field if it contains floating-point or decimal values
        probability_type = probabilities["probability"].dtype
        if probability_type.is_float() or probability_type.is_decimal():
            probability_uint8 = pl.Series(
                probabilities.select(
                    pl.col("probability").mul(100).round(0).cast(pl.UInt8)
                )
            )

            # Check max value only if the table is not empty
            max_prob = probability_uint8.max()
            if max_prob is not None and max_prob > 100:
                p_max = max_prob
                p_min = probability_uint8.min()
                raise ValueError(f"Probability range misconfigured: [{p_min}, {p_max}]")

            probabilities = probabilities.replace_column(
                probabilities.get_column_index("probability"), probability_uint8
            )

        # need schema in format recognised by polars
        self.probabilities = probabilities.cast(pl.Schema(SCHEMA_RESULTS))

    @property
    def clusters(self) -> pl.DataFrame:
        """Retrieve new clusters implied by these results."""
        if self._clusters is None:
            im = IntMap()
            self._clusters = to_clusters(
                results=self.probabilities, dtype=pl.Int64, hash_func=im.index
            )
        return self._clusters

    def root_leaf(self) -> pl.DataFrame:
        """Returns all roots and leaves implied by these results."""
        if self.left_root_leaf is None:
            raise RuntimeError(
                "This Results object wasn't instantiated for validation features."
            )

        parents_root_leaf = self.left_root_leaf.select(["id", "leaf_id"])
        if self.right_root_leaf is not None:
            parents_root_leaf = pl.concat(
                [
                    parents_root_leaf,
                    self.right_root_leaf.select(["id", "leaf_id"]),
                ]
            )

        # Go from parent-child (where child could be the root of another model)
        # to root-leaf, where leaf is a source cluster ID
        root_leaf_res = (
            self.clusters.rename({"parent": "root_id"})
            .join(parents_root_leaf, left_on="child", right_on="id")
            .select(["root_id", "leaf_id"])
            .unique()
        )

        # Generate root-leaf for those input rows that weren't merged by this model
        unmerged_ids_rows = (
            parents_root_leaf.select("id", "leaf_id")
            .join(
                self.clusters.select("child"),
                left_on="id",
                right_on="child",
                how="anti",
            )
            .rename({"id": "root_id"})
            .select(["root_id", "leaf_id"])
            .unique()
        )

        return pl.concat([root_leaf_res, unmerged_ids_rows])



"""Client-side helpers for retrieving and preparing evaluation samples."""

from typing import cast

import polars as pl
from pydantic import BaseModel
from sqlalchemy.exc import OperationalError

from matchbox.client import _handler
from matchbox.client.dags import DAG
from matchbox.client.results import Results
from matchbox.common.dtos import ModelResolutionName, ModelResolutionPath, SourceConfig
from matchbox.common.eval import Judgement, ModelComparison, precision_recall
from matchbox.common.exceptions import MatchboxSourceTableError


class EvaluationItem(BaseModel):
    """A cluster awaiting evaluation, with pre-computed display data."""

    model_config = {"arbitrary_types_allowed": True}

    cluster_id: int
    dataframe: pl.DataFrame  # Original raw data (needed for judgement leaf IDs)
    display_data: dict[str, list[str]]  # field_name -> [val1, val2, val3]
    duplicate_groups: list[list[int]]  # Groups of leaf_ids with identical data
    display_columns: list[int]  # Representative leaf_id for each displayed column
    assignments: dict[int, str] = {}  # display_column_index -> group_letter


def create_judgement(item: EvaluationItem, user_id: int) -> Judgement:
    """Convert item assignments to Judgement - no default group assignment.

    Args:
        item: Evaluation item with assignments
        user_id: User ID for the judgement

    Returns:
        Judgement with endorsed groups based on assignments
    """
    groups: dict[str, list[int]] = {}

    for col_idx, group in item.assignments.items():
        leaf_ids = item.duplicate_groups[col_idx]
        groups.setdefault(group, []).extend(leaf_ids)

    endorsed = [sorted(set(leaf_ids)) for leaf_ids in groups.values()]
    return Judgement(user_id=user_id, shown=item.cluster_id, endorsed=endorsed)


def create_evaluation_item(
    df: pl.DataFrame, source_configs: list[tuple[str, SourceConfig]], cluster_id: int
) -> EvaluationItem:
    """Create EvaluationItem with pre-computed display data."""
    # Get all data columns (exclude metadata columns)
    data_cols = [c for c in df.columns if c not in ["root", "leaf", "key"]]

    # Extract field names from source configs
    field_names = []
    for _, config in source_configs:
        for field in config.index_fields:
            if field.name not in field_names:
                field_names.append(field.name)

    if not data_cols:
        # No data columns found - return empty item
        return EvaluationItem(
            cluster_id=cluster_id,
            dataframe=df,
            display_data={},
            duplicate_groups=[],
            display_columns=[],
            assignments={},
        )

    # Deduplicate using polars group_by
    df_dedup = df.select(["leaf"] + data_cols)
    grouped = df_dedup.group_by(data_cols, maintain_order=True).agg(pl.col("leaf"))

    duplicate_groups = [group for group in grouped["leaf"]]
    display_columns = [group[0] for group in duplicate_groups]

    # Build display data
    display_data = {}
    for field_name in field_names:
        qualified_cols = [c for c in data_cols if c.endswith(f"_{field_name}")]

        values = []
        for leaf_id in display_columns:
            row = df.filter(pl.col("leaf") == leaf_id).row(0, named=True)
            val = next(
                (str(row.get(c, "")).strip() for c in qualified_cols if row.get(c)), ""
            )
            values.append(val)

        if any(values):  # Only include fields with data
            display_data[field_name] = values

    return EvaluationItem(
        cluster_id=cluster_id,
        dataframe=df,
        display_data=display_data,
        duplicate_groups=duplicate_groups,
        display_columns=display_columns,
        assignments={},
    )


def get_samples(
    n: int,
    dag: DAG,
    user_id: int,
    resolution: ModelResolutionName | None = None,
) -> dict[int, EvaluationItem]:
    """Retrieve samples enriched with source data as EvaluationItems.

    Args:
        n: Number of clusters to sample
        dag: DAG for which to retrieve samples
        user_id: ID of the user requesting the samples
        resolution: The optional resolution from which to sample. If not provided,
            the final step in the DAG is used

    Returns:
        Dictionary of cluster ID to EvaluationItems describing the cluster

    Raises:
        MatchboxSourceTableError: If a source cannot be queried from a location using
            provided or default clients.
    """
    if resolution:
        resolution_path: ModelResolutionPath = dag.get_model(resolution).resolution_path
    else:
        resolution_path: ModelResolutionPath = dag.final_step.resolution_path

    samples: pl.DataFrame = cast(
        pl.DataFrame,
        pl.from_arrow(
            _handler.sample_for_eval(n=n, resolution=resolution_path, user_id=user_id)
        ),
    )

    if not len(samples):
        return {}

    results_by_source: list[pl.DataFrame] = []
    source_configs: list[tuple[str, SourceConfig]] = []

    for source_resolution in samples["source"].unique():
        try:
            source = dag.get_source(source_resolution)
        except ValueError as exc:  # pragma: no cover - defensive path
            raise MatchboxSourceTableError(
                f"Source '{source_resolution}' not found in DAG. "
                "Ensure DAG was loaded with all resolutions."
            ) from exc

        source_configs.append((source_resolution, source.config))

        samples_by_source = samples.filter(pl.col("source") == source_resolution)
        keys_by_source = samples_by_source["key"].to_list()

        try:
            source_data = pl.concat(
                source.fetch(batch_size=10_000, qualify_names=True, keys=keys_by_source)
            )
        except OperationalError as exc:
            raise MatchboxSourceTableError(
                f"Could not query source '{source_resolution}' from warehouse. "
                "Check warehouse connection and ensure source table exists."
            ) from exc

        samples_and_source = samples_by_source.join(
            source_data, left_on="key", right_on=source.qualified_key
        )
        desired_columns = ["root", "leaf", "key"] + source.qualified_index_fields
        results_by_source.append(samples_and_source[desired_columns])

    if not results_by_source:
        return {}

    all_results: pl.DataFrame = pl.concat(results_by_source, how="diagonal")

    results_by_root: dict[int, EvaluationItem] = {}
    for root in all_results["root"].unique():
        cluster_df = all_results.filter(pl.col("root") == root).drop("root")
        evaluation_item = create_evaluation_item(cluster_df, source_configs, root)
        results_by_root[root] = evaluation_item

    return results_by_root


class EvalData:
    """Object which caches evaluation data to measure model performance."""

    def __init__(self) -> None:
        """Download judgement and expansion data used to compute evaluation metrics."""
        self.judgements, self.expansion = _handler.download_eval_data()

    def precision_recall(
        self, results: Results, threshold: float
    ) -> tuple[float, float]:
        """Compute precision and recall for a given Results object."""
        if not len(results.clusters):
            raise ValueError("No clusters suggested by these results.")

        threshold = int(threshold * 100)
        root_leaf = results.root_leaf().rename({"root_id": "root", "leaf_id": "leaf"})
        values = precision_recall([root_leaf], self.judgements, self.expansion)[0]
        return values[0], values[1]


def compare_models(resolutions: list[ModelResolutionPath]) -> ModelComparison:
    """Compare metrics of models based on cached evaluation data."""
    return _handler.compare_models(resolutions)



"""Add upload stage to resolutions.

Revision ID: c774fd4b69f8
Revises: 8c7f757b1046
Create Date: 2025-10-31 07:42:00.669155

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = "c774fd4b69f8"
down_revision: str | None = "8c7f757b1046"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None

upload_stages = sa.Enum(
    "READY", "PROCESSING", "COMPLETE", name="upload_stages", schema="mb"
)


def upgrade() -> None:
    """Upgrade schema."""
    # Create upload stage
    connection = op.get_bind()
    upload_stages.create(connection)
    op.add_column(
        "resolutions",
        sa.Column("upload_stage", upload_stages, nullable=True),
        schema="mb",
    )

    connection.execute(
        sa.text("UPDATE mb.resolutions SET upload_stage = :stage"),
        {"stage": "COMPLETE"},
    )

    op.alter_column("resolutions", "upload_stage", nullable=False, schema="mb")

    # Change "content hash" to "fingerprint"
    op.execute("UPDATE mb.resolutions SET hash = decode('', 'hex') WHERE hash IS NULL")

    op.alter_column(
        "resolutions",
        "hash",
        new_column_name="fingerprint",
        schema="mb",
        existing_type=postgresql.BYTEA(),
        nullable=False,
    )


def downgrade() -> None:
    """Downgrade schema."""
    # Drop upload stage
    op.drop_column("resolutions", "upload_stage", schema="mb")
    upload_stages.drop(op.get_bind())

    # Change "fingerprint" to "content hash"
    op.alter_column(
        "resolutions",
        "fingerprint",
        new_column_name="hash",
        schema="mb",
        existing_type=postgresql.BYTEA(),
        nullable=True,
    )



"""API routes for the Matchbox server."""

from collections.abc import Awaitable, Callable
from importlib.metadata import version
from pathlib import Path
from typing import Annotated

from fastapi import (
    Depends,
    FastAPI,
    HTTPException,
    Query,
    Request,
    Response,
)
from fastapi.encoders import jsonable_encoder
from fastapi.openapi.docs import (
    get_swagger_ui_html,
    get_swagger_ui_oauth2_redirect_html,
)
from fastapi.responses import JSONResponse
from fastapi.staticfiles import StaticFiles
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.responses import HTMLResponse

from matchbox.common.arrow import table_to_buffer
from matchbox.common.dtos import (
    BackendCountableType,
    BackendResourceType,
    CollectionName,
    CountResult,
    CRUDOperation,
    LoginAttempt,
    LoginResult,
    Match,
    ModelResolutionName,
    NotFoundError,
    OKMessage,
    ResolutionPath,
    ResourceOperationStatus,
    RunID,
    SourceResolutionName,
    SourceResolutionPath,
)
from matchbox.common.exceptions import (
    MatchboxCollectionNotFoundError,
    MatchboxDeletionNotConfirmed,
    MatchboxResolutionNotFoundError,
    MatchboxRunNotFoundError,
)
from matchbox.server.api.dependencies import (
    BackendDependency,
    ParquetResponse,
    authorisation_dependencies,
    lifespan,
)
from matchbox.server.api.routers import collection, eval

app = FastAPI(
    title="matchbox API",
    version=version("matchbox_db"),
    lifespan=lifespan,
    docs_url=None,
    redoc_url=None,
)
app.include_router(collection.router)
app.include_router(eval.router)

static_dir = Path(__file__).parent / "static"
app.mount("/static", StaticFiles(directory=static_dir), name="static")


@app.exception_handler(StarletteHTTPException)
async def http_exception_handler(
    request: Request, exc: StarletteHTTPException
) -> JSONResponse:
    """Overwrite the default JSON schema for an `HTTPException`."""
    return JSONResponse(
        content=jsonable_encoder(exc.detail), status_code=exc.status_code
    )


@app.exception_handler(MatchboxCollectionNotFoundError)
async def collection_not_found_handler(
    request: Request, exc: MatchboxCollectionNotFoundError
) -> JSONResponse:
    """Handle collection not found errors."""
    detail = NotFoundError(
        details=str(exc), entity=BackendResourceType.COLLECTION
    ).model_dump()
    return JSONResponse(
        status_code=404,
        content=jsonable_encoder(detail),
    )


@app.exception_handler(MatchboxRunNotFoundError)
async def run_not_found_handler(
    request: Request, exc: MatchboxRunNotFoundError
) -> JSONResponse:
    """Handle run not found errors."""
    detail = NotFoundError(
        details=str(exc), entity=BackendResourceType.RUN
    ).model_dump()
    return JSONResponse(
        status_code=404,
        content=jsonable_encoder(detail),
    )


@app.exception_handler(MatchboxResolutionNotFoundError)
async def resolution_not_found_handler(
    request: Request, exc: MatchboxResolutionNotFoundError
) -> JSONResponse:
    """Handle resolution not found errors."""
    detail = NotFoundError(
        details=str(exc), entity=BackendResourceType.RESOLUTION
    ).model_dump()
    return JSONResponse(
        status_code=404,
        content=jsonable_encoder(detail),
    )


@app.exception_handler(MatchboxDeletionNotConfirmed)
async def deletion_not_confirmed_handler(
    request: Request, exc: MatchboxDeletionNotConfirmed
) -> JSONResponse:
    """Handle deletion not confirmed errors."""
    # Extract resource name from request
    path_parts = request.url.path.split("/")
    resource_name = path_parts[-1] if path_parts else "Unknown"

    detail = ResourceOperationStatus(
        success=False,
        target=resource_name,
        operation=CRUDOperation.DELETE,
        details=str(exc),
    ).model_dump()

    return JSONResponse(
        status_code=409,
        content=jsonable_encoder(detail),
    )


@app.middleware("http")
async def add_security_headers(
    request: Request, call_next: Callable[[Request], Awaitable[Response]]
) -> Response:
    """Improve security by adding headers to all responses."""
    response: Response = await call_next(request)
    response.headers["Cache-control"] = "no-store, no-cache"
    response.headers["Content-Security-Policy"] = (
        # Restrict by default
        "default-src 'none'; frame-ancestors 'none'; form-action 'none';"
        # Load Swagger CSS, favicon and openapi.json
        "style-src 'self'; img-src 'self' data:; connect-src 'self'; "
        # Load Swagger JS, hard-coding the expected file hash
        "script-src 'self' 'sha256-QOOQu4W1oxGqd2nbXbxiA1Di6OHQOLQD+o+G9oWL8YY='"
    )
    response.headers["Strict-Transport-Security"] = (
        "max-age=31536000; includeSubDomains"
    )
    response.headers["X-Frame-Options"] = "DENY"
    response.headers["X-Content-Type-Options"] = "nosniff"
    return response


@app.get("/health")
async def healthcheck() -> OKMessage:
    """Perform a health check and return the status."""
    return OKMessage()


@app.post(
    "/login",
)
def login(
    backend: BackendDependency,
    credentials: LoginAttempt,
) -> LoginResult:
    """Receives a user name and returns a user ID."""
    return LoginResult(user_id=backend.login(credentials.user_name))


# Retrieval


@app.get(
    "/query",
    responses={404: {"model": NotFoundError}},
)
def query(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    source: SourceResolutionName,
    return_leaf_id: bool,
    resolution: ModelResolutionName | None = None,
    threshold: int | None = None,
    limit: int | None = None,
) -> ParquetResponse:
    """Query Matchbox for matches based on a source resolution name."""
    try:
        res = backend.query(
            source=SourceResolutionPath(
                collection=collection,
                run=run_id,
                name=source,
            ),
            point_of_truth=ResolutionPath(
                collection=collection, run=run_id, name=resolution
            )
            if resolution
            else None,
            threshold=threshold,
            return_leaf_id=return_leaf_id,
            limit=limit,
        )
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendResourceType.RESOLUTION
            ).model_dump(),
        ) from e

    buffer = table_to_buffer(res)
    return ParquetResponse(buffer.getvalue())


@app.get(
    "/match",
    responses={404: {"model": NotFoundError}},
)
def match(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    targets: Annotated[list[SourceResolutionName], Query()],
    source: SourceResolutionName,
    key: str,
    resolution: ModelResolutionName,
    threshold: int | None = None,
) -> list[Match]:
    """Match a source key against a list of target source resolutions."""
    targets = [
        SourceResolutionPath(collection=collection, run=run_id, name=t) for t in targets
    ]
    try:
        res = backend.match(
            key=key,
            source=SourceResolutionPath(
                collection=collection,
                run=run_id,
                name=source,
            ),
            targets=targets,
            point_of_truth=ResolutionPath(
                collection=collection,
                run=run_id,
                name=resolution,
            ),
            threshold=threshold,
        )
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendResourceType.RESOLUTION
            ).model_dump(),
        ) from e

    return res


# Admin


@app.get("/database/count")
def count_backend_items(
    backend: BackendDependency,
    entity: BackendCountableType | None = None,
) -> CountResult:
    """Count the number of various entities in the backend."""

    def get_count(e: BackendCountableType) -> int:
        return getattr(backend, str(e)).count()

    if entity is not None:
        return CountResult(entities={str(entity): get_count(entity)})
    else:
        res = {str(e): get_count(e) for e in BackendCountableType}
        return CountResult(entities=res)


@app.delete(
    "/database",
    responses={
        409: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
        500: {
            "model": ResourceOperationStatus,
            **ResourceOperationStatus.error_examples(),
        },
    },
    dependencies=[Depends(authorisation_dependencies)],
)
def clear_database(
    backend: BackendDependency,
    certain: Annotated[
        bool,
        Query(
            description=(
                "Confirm deletion of all data in the database whilst retaining tables"
            )
        ),
    ] = False,
) -> ResourceOperationStatus:
    """Delete all data from the backend whilst retaining tables."""
    try:
        backend.clear(certain=certain)
        return ResourceOperationStatus(
            success=True, target="Database", operation=CRUDOperation.DELETE
        )
    except MatchboxDeletionNotConfirmed:
        raise
    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=ResourceOperationStatus(
                success=False,
                target="Database",
                operation=CRUDOperation.DELETE,
                details=str(e),
            ),
        ) from e


# Swagger UI


@app.get("/docs", include_in_schema=False)
async def custom_swagger_ui_html() -> HTMLResponse:
    """Get locally hosted docs."""
    return get_swagger_ui_html(
        openapi_url=app.openapi_url,
        title=app.title + " - Swagger UI",
        oauth2_redirect_url=app.swagger_ui_oauth2_redirect_url,
        swagger_js_url="/static/swagger-ui-bundle.js",
        swagger_css_url="/static/swagger-ui.css",
        swagger_favicon_url="/static/favicon.png",
    )


@app.get(app.swagger_ui_oauth2_redirect_url, include_in_schema=False)
async def swagger_ui_redirect() -> HTMLResponse:
    """Helper for OAuth2."""
    return get_swagger_ui_oauth2_redirect_html()



"""Matchbox PostgreSQL database connection."""

from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import Any

from adbc_driver_postgresql import dbapi as adbc_dbapi
from alembic import command
from alembic.config import Config
from pydantic import BaseModel, Field
from sqlalchemy import (
    URL,
    Engine,
    MetaData,
    Table,
    create_engine,
    inspect,
    text,
)
from sqlalchemy.orm import Session, declarative_base, sessionmaker
from sqlalchemy.pool import PoolProxiedConnection, QueuePool

from matchbox.common.logging import logger
from matchbox.server.base import MatchboxBackends, MatchboxServerSettings


class MatchboxPostgresCoreSettings(BaseModel):
    """PostgreSQL-specific settings for Matchbox."""

    host: str
    port: int
    user: str
    password: str
    database: str
    db_schema: str
    alembic_config: Path = Field(
        default=Path("src/matchbox/server/postgresql/alembic.ini")
    )

    def get_alembic_config(self) -> Config:
        """Get the Alembic config."""
        config = Config(self.alembic_config)
        db_url = URL.create(
            "postgresql+psycopg",
            username=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.database,
        )
        config.set_main_option(
            "sqlalchemy.url", db_url.render_as_string(hide_password=False)
        )
        return config


class MatchboxPostgresSettings(MatchboxServerSettings):
    """Settings for the Matchbox PostgreSQL backend.

    Inherits the core settings and adds the PostgreSQL-specific settings.
    """

    backend_type: MatchboxBackends = MatchboxBackends.POSTGRES

    postgres: MatchboxPostgresCoreSettings = Field(
        default_factory=MatchboxPostgresCoreSettings
    )


class MatchboxDatabase:
    """Matchbox PostgreSQL database connection."""

    def __init__(self, settings: MatchboxPostgresSettings) -> None:
        """Initialise the database connection."""
        self.settings = settings
        self._engine: Engine | None = None
        self._SessionLocal: sessionmaker | None = None
        self._adbc_pool: QueuePool | None = None
        self._source_adbc_connection: adbc_dbapi.Connection | None = None
        self.MatchboxBase = declarative_base(
            metadata=MetaData(schema=settings.postgres.db_schema)
        )
        self.alembic_config = settings.postgres.get_alembic_config()

    def connection_string(self, driver: bool = True) -> str:
        """Get the connection string for PostgreSQL."""
        driver_string = ""
        if driver:
            driver_string = "+psycopg"
        return (
            f"postgresql{driver_string}://{self.settings.postgres.user}:{self.settings.postgres.password}"
            f"@{self.settings.postgres.host}:{self.settings.postgres.port}/"
            f"{self.settings.postgres.database}"
        )

    def _connect(self) -> None:
        """Connect to the database."""
        self._engine = create_engine(
            url=self.connection_string(), logging_name="matchbox.engine", echo=False
        )
        self._SessionLocal = sessionmaker(
            autocommit=False, autoflush=False, bind=self._engine
        )

    def _disconnect(self) -> None:
        if self._engine:
            self._engine.dispose()
            self._engine = None
            self._SessionLocal = None

    def _connect_adbc(self) -> None:
        self._source_adbc_connection = adbc_dbapi.connect(
            self.connection_string(driver=False)
        )
        self._adbc_pool = QueuePool(
            self._source_adbc_connection.adbc_clone,
        )

    def _disconnect_adbc(self) -> None:
        if self._adbc_pool:
            self._adbc_pool.dispose()
            self._adbc_pool = None

            self._source_adbc_connection.close()
            self._source_adbc_connection = None

    def _reset_connections(self) -> None:
        """Dispose and re-initialise SQLAlchemy and ADBC connection managers."""
        self._disconnect()
        self._connect()

        self._disconnect_adbc()
        self._connect_adbc()

    def get_engine(self) -> Engine:
        """Get the database engine."""
        if not self._engine:
            self._connect()
        return self._engine

    def get_session(self) -> Session:
        """Get a new session."""
        if not self._SessionLocal:
            self._connect()
        return self._SessionLocal()

    @contextmanager
    def get_adbc_connection(self) -> Generator[PoolProxiedConnection, Any, Any]:
        """Get a new ADBC connection wrapped by a SQLAlchemy pool proxy.

        The connection must be used within a context manager.
        """
        if not self._adbc_pool:
            self._connect_adbc()

        conn = self._adbc_pool.connect()
        try:
            yield conn
        finally:
            conn.close()

    def run_migrations(self) -> None:
        """Create the database and all tables expected in the schema."""
        alembic_version = self._look_for_alembic_version()
        engine = self.get_engine()
        logger.info("Determinded alembic in use so upgrading to head")
        if alembic_version is not None:
            command.upgrade(self.alembic_config, "head")
        else:
            logger.info(
                "Determinded alembic not in use so dropping schema if it "
                "exists prior to upgrading to head. "
            )
            with engine.connect() as conn:
                conn.execute(text("DROP SCHEMA IF EXISTS mb CASCADE;"))
                conn.commit()
            command.upgrade(self.alembic_config, "head")

    def clear_database(self) -> None:
        """Delete all rows in every table in the database schema.

        - TRUNCATE tables that are part of the core ORM (preserves structure)
        - DROP tables that are not in the ORM (removes temporary/test tables)
        """
        schema_name = self.MatchboxBase.metadata.schema
        engine = self.get_engine()  # Get the engine

        with self.get_session() as session:
            try:
                # Get all tables that actually exist in the database
                discovered_metadata = MetaData(schema=schema_name)
                discovered_metadata.reflect(
                    bind=engine, schema=schema_name
                )  # Use engine here

                # Get table names from the core ORM
                orm_table_names = {
                    table.name for table in self.MatchboxBase.metadata.tables.values()
                }

                # Categorise discovered tables
                tables_to_truncate: list[Table] = []
                tables_to_drop: list[Table] = []

                for table in discovered_metadata.tables.values():
                    if table.name == "alembic_version":
                        # Alembic version is in public but just in case
                        continue
                    elif table.name in orm_table_names:
                        tables_to_truncate.append(table)
                    else:
                        tables_to_drop.append(table)

                # TRUNCATE core ORM tables (preserves schema structure)
                if tables_to_truncate:
                    for table in tables_to_truncate:
                        session.execute(
                            text(f"TRUNCATE TABLE {table.fullname} CASCADE")
                        )

                # DROP temporary/test tables (removes them completely)
                if tables_to_drop:
                    for table in tables_to_drop:
                        session.execute(text(f"DROP TABLE {table.fullname} CASCADE"))

                session.commit()

            except Exception as e:
                logger.error(f"Error clearing database: {e}")
                session.rollback()
                raise

    def drop_database(self) -> None:
        """Drop all tables in the database schema and re-recreate them."""
        command.downgrade(self.alembic_config, "base")
        command.upgrade(self.alembic_config, "head")

    def vacuum_analyze(self, *table_names: str) -> None:
        """Run VACUUM ANALYZE on specified tables.

        VACUUM ANALYZE reclaims storage and updates statistics for the query planner.
        PostgreSQL may not fully utilise indexes until VACUUM ANALYZE is run.
        According to https://www.postgresql.org/docs/current/sql-vacuum.html,
        VACUUM ANALYZE is recommended over just ANALYZE for optimal performance.

        Args:
            *table_names: Fully qualified table names to vacuum. If none provided,
                vacuums the entire database.
        """
        engine = self.get_engine()
        with engine.connect().execution_options(isolation_level="AUTOCOMMIT") as conn:
            if table_names:
                for table_name in table_names:
                    conn.execute(text(f"VACUUM ANALYZE {table_name};"))
            else:
                conn.execute(text("VACUUM ANALYZE;"))

    def _look_for_alembic_version(self) -> bool:
        engine = self.get_engine()
        inspector = inspect(engine)
        alembic_version_table = "alembic_version" in inspector.get_table_names(
            schema="public"
        )
        if alembic_version_table:
            with engine.connect() as conn:
                result = conn.execute(
                    text("SELECT version_num FROM public.alembic_version;")
                )
                alembic_version = result.scalar()
        else:
            alembic_version = None
        return alembic_version

    @property
    def sorted_tables(self) -> list[Table]:
        """Return a list of SQLAlchemy tables in order of creation."""
        return self.MatchboxBase.metadata.sorted_tables


# Global database instance -- everything should use this

MBDB = MatchboxDatabase(MatchboxPostgresSettings())



import json

import pyarrow as pa
import pytest
from httpx import Response
from respx.router import MockRouter
from sqlalchemy import Engine

from matchbox.client.dags import DAG
from matchbox.client.models import Model, add_model_class
from matchbox.client.models.linkers.base import LinkerSettings
from matchbox.client.queries import Query
from matchbox.common.arrow import table_to_buffer
from matchbox.common.dtos import (
    BackendResourceType,
    CRUDOperation,
    ModelConfig,
    ModelType,
    NotFoundError,
    Resolution,
    ResourceOperationStatus,
    UploadInfo,
    UploadStage,
)
from matchbox.common.exceptions import (
    MatchboxDeletionNotConfirmed,
    MatchboxServerFileError,
)
from matchbox.common.factories.models import MockLinker, model_factory
from matchbox.common.factories.sources import (
    source_factory,
)


def test_init_and_run_model(sqlite_warehouse: Engine, matchbox_api: MockRouter) -> None:
    """Test that model can be initialised and run correctly."""
    # Register "custom" model
    add_model_class(MockLinker)

    # Mock API
    foo = source_factory(engine=sqlite_warehouse).write_to_location()
    bar = source_factory(engine=sqlite_warehouse).write_to_location()

    foo_leafy_data = foo.data.append_column(
        "leaf_id", pa.array(range(len(foo.data)), type=pa.int64())
    )
    bar_leafy_data = foo.data.append_column(
        "leaf_id",
        pa.array(
            range(len(foo_leafy_data), len(foo_leafy_data) + len(bar.data)),
            type=pa.int64(),
        ),
    )

    # Mock API
    query_endpoint = matchbox_api.get("/query").mock(
        side_effect=[
            # First query
            Response(200, content=table_to_buffer(foo.data).read()),
            Response(200, content=table_to_buffer(bar.data).read()),
            # Second query (cached)
            Response(200, content=table_to_buffer(foo.data).read()),
            Response(200, content=table_to_buffer(bar.data).read()),
            # Third query (for validation)
            Response(200, content=table_to_buffer(foo_leafy_data).read()),
            Response(200, content=table_to_buffer(bar_leafy_data).read()),
        ]
    )
    dag = DAG("collection")
    foo_query = Query(foo.source, dag=dag)
    bar_query = Query(bar.source, dag=dag)

    model = Model(
        dag=dag,
        name="name",
        description="description",
        model_class=MockLinker,
        model_settings=LinkerSettings(),
        left_query=foo_query,
        right_query=bar_query,
    )

    assert model.config == ModelConfig(
        type=ModelType.LINKER,
        model_class="MockLinker",
        model_settings=json.dumps({"left_id": "l.field", "right_id": "r.field"}),
        left_query=foo_query.config,
        right_query=bar_query.config,
    )

    model.run()
    assert model.results.left_root_leaf is None
    assert model.results.right_root_leaf is None

    # When caching queries, the first ones still have to go through
    old_query_count = query_endpoint.call_count
    model.run(cache_queries=True)
    assert query_endpoint.call_count > old_query_count
    # Subsequent queries are cached
    old_query_count = query_endpoint.call_count
    model.run(cache_queries=True)
    assert query_endpoint.call_count == old_query_count

    model.run(for_validation=True)
    assert model.results.left_root_leaf is not None
    assert model.results.right_root_leaf is not None


def test_model_sync(matchbox_api: MockRouter) -> None:
    # Mock model
    testkit = model_factory(model_type="linker")

    # Mock the routes:
    # Resolution doesn't yet exist
    matchbox_api.get(
        f"/collections/{testkit.model.dag.name}/runs/{testkit.model.dag.run}/resolutions/{testkit.model.name}"
    ).mock(
        return_value=Response(
            404,
            json=NotFoundError(
                details="Model not found", entity=BackendResourceType.RESOLUTION
            ).model_dump(),
        )
    )

    # Resolution can be inserted
    insert_config_route = matchbox_api.post(
        f"/collections/{testkit.model.dag.name}/runs/{testkit.model.dag.run}/resolutions/{testkit.model.name}"
    ).mock(
        return_value=Response(
            201,
            content=ResourceOperationStatus(
                success=True,
                target=f"Resolution {testkit.model.name}",
                operation=CRUDOperation.CREATE,
            ).model_dump_json(),
        )
    )

    # Resolution data can be inserted
    insert_results_route = matchbox_api.post(
        f"/collections/{testkit.model.dag.name}/runs/{testkit.model.dag.run}/resolutions/{testkit.model.name}/data"
    ).mock(
        return_value=Response(
            202,
            json=ResourceOperationStatus(
                success=True, target="", operation=CRUDOperation.CREATE
            ).model_dump(),
        )
    )

    # Later, resolution can be updated
    update_route = matchbox_api.put(
        f"/collections/{testkit.model.dag.name}/runs/{testkit.model.dag.run}/resolutions/{testkit.model.name}"
    ).mock(
        return_value=Response(
            200,
            content=ResourceOperationStatus(
                success=True,
                target=f"Resolution {testkit.model.name}",
                operation=CRUDOperation.UPDATE,
            ).model_dump_json(),
        )
    )

    # Later, resolution can be deleted and recreated
    delete_route = matchbox_api.delete(
        f"/collections/{testkit.model.dag.name}/runs/{testkit.model.dag.run}/resolutions/{testkit.model.name}"
    ).mock(
        return_value=Response(
            200,
            content=ResourceOperationStatus(
                success=True,
                target=f"Resolution {testkit.model.name}",
                operation=CRUDOperation.DELETE,
            ).model_dump_json(),
        )
    )

    # -- ERRORS --

    # Can't sync before running
    with pytest.raises(RuntimeError, match="must be run"):
        testkit.model.sync()

    # We now run, but test that upload failure is handled
    testkit.fake_run()
    # Before and after upload, server accepts data - the second time reporting error
    matchbox_api.get(
        f"/collections/{testkit.model.dag.name}/runs/{testkit.model.dag.run}/resolutions/{testkit.model.name}/data/status"
    ).mock(
        side_effect=[
            Response(200, json=UploadInfo(stage=UploadStage.READY).model_dump()),
            Response(
                200,
                json=UploadInfo(stage=UploadStage.READY, error="error").model_dump(),
            ),
        ]
    )
    with pytest.raises(MatchboxServerFileError, match="error"):
        testkit.model.sync()

    # -- FIRST TIME INSERTION --

    # Before upload, resolution is ready for data, after it is complete
    matchbox_api.get(
        f"/collections/{testkit.model.dag.name}/runs/{testkit.model.dag.run}/resolutions/{testkit.model.name}/data/status"
    ).mock(
        side_effect=[
            Response(200, json=UploadInfo(stage=UploadStage.READY).model_dump()),
            Response(200, json=UploadInfo(stage=UploadStage.COMPLETE).model_dump()),
        ]
    )

    # Sync the source, successfully
    testkit.model.sync()
    # Source was created, not updated or deleted
    assert insert_config_route.called
    assert not update_route.called
    assert not delete_route.called
    # Resolution metadata was correct
    resolution_call = Resolution.model_validate_json(
        insert_config_route.calls.last.request.content.decode("utf-8")
    )
    assert resolution_call == testkit.model.to_resolution()
    # Resolution data was correct
    assert (
        b"Content-Disposition: form-data;"
        in insert_results_route.calls.last.request.content
    )
    assert b"PAR1" in insert_results_route.calls.last.request.content

    # -- SOFT UPDATE --

    insert_results_route.reset()
    # Mock endpoint now returns existing resolution
    matchbox_api.get(
        f"/collections/{testkit.model.dag.name}/runs/{testkit.model.dag.run}/resolutions/{testkit.model.name}"
    ).mock(return_value=Response(200, json=testkit.model.to_resolution().model_dump()))

    # Mock endpoint now declares data is present already
    matchbox_api.get(
        f"/collections/{testkit.model.dag.name}/runs/{testkit.model.dag.run}/resolutions/{testkit.model.name}/data/status"
    ).mock(
        return_value=Response(
            200, json=UploadInfo(stage=UploadStage.COMPLETE).model_dump()
        )
    )

    testkit.model.sync()
    # Resolution was compatible: ensure it was updated, not deleted
    assert update_route.called
    assert not delete_route.called
    # The data did not need to be updated
    assert not insert_results_route.called

    # -- HARD UPDATE --

    insert_results_route.reset()
    # Changing data requires deletion and re-insertion
    testkit.probabilities = testkit.probabilities.slice(1, 3)
    testkit.fake_run()
    # Resolution data is first ready to upload, and then uploaded
    matchbox_api.get(
        f"/collections/{testkit.model.dag.name}/runs/{testkit.model.dag.run}/resolutions/{testkit.model.name}/data/status"
    ).mock(
        side_effect=[
            Response(200, json=UploadInfo(stage=UploadStage.READY).model_dump()),
            Response(200, json=UploadInfo(stage=UploadStage.COMPLETE).model_dump()),
        ]
    )

    testkit.model.sync()
    assert delete_route.called
    assert insert_results_route.called


def test_truth_getter() -> None:
    """Test getting model truth threshold from config."""
    # Create testkit with specific truth value
    testkit = model_factory(model_type="linker")
    # Update the model to have a truth value
    testkit.model._truth = 90  # Integer truth value (90 = 0.9 as float)

    # Get truth as float
    truth = testkit.model.truth

    # Verify it returns the correct value converted to float
    assert truth == 0.9


def test_truth_setter_validation_error() -> None:
    """Test setting invalid truth values."""
    testkit = model_factory(model_type="linker")

    # Attempt to set an invalid truth value using the validated setter
    with pytest.raises(ValueError):
        testkit.model.truth = 1.5


def test_delete_resolution(matchbox_api: MockRouter) -> None:
    """Test successfully deleting a resolution."""
    # Create test model using factory
    testkit = model_factory()

    # Mock the DELETE endpoint with success response
    route = matchbox_api.delete(
        f"/collections/{testkit.model.dag.name}/runs/{testkit.model.dag.run}/resolutions/{testkit.model.name}",
        params={"certain": True},
    ).mock(
        return_value=Response(
            200,
            json=ResourceOperationStatus(
                success=True,
                target=f"Resolution {testkit.model.name}",
                operation=CRUDOperation.DELETE,
            ).model_dump(),
        )
    )

    # Delete the model
    response = testkit.model.delete(certain=True)

    # Verify the response and API call
    assert response
    assert route.called
    assert route.calls.last.request.url.params["certain"] == "true"


def test_delete_resolution_needs_confirmation(matchbox_api: MockRouter) -> None:
    """Test attempting to delete a resolution without confirmation returns 409."""
    # Create test model using factory
    testkit = model_factory()

    # Mock the DELETE endpoint with 409 confirmation required response
    error_details = "Cannot delete model with dependent models: dedupe1, dedupe2"
    route = matchbox_api.delete(
        f"/collections/{testkit.model.dag.name}/runs/{testkit.model.dag.run}/resolutions/{testkit.model.name}"
    ).mock(
        return_value=Response(
            409,
            json=ResourceOperationStatus(
                success=False,
                target=f"Resolution {testkit.model.name}",
                operation=CRUDOperation.DELETE,
                details=error_details,
            ).model_dump(),
        )
    )

    # Attempt to delete without certain=True
    with pytest.raises(MatchboxDeletionNotConfirmed):
        testkit.model.delete()

    # Verify the response and API call
    assert route.called
    assert route.calls.last.request.url.params["certain"] == "false"



from unittest.mock import Mock, patch

import pytest
from fastapi.testclient import TestClient

from matchbox.common.arrow import table_to_buffer
from matchbox.common.dtos import (
    BackendResourceType,
    CRUDOperation,
    NotFoundError,
    ResolutionPath,
    ResourceOperationStatus,
    UploadInfo,
    UploadStage,
)
from matchbox.common.exceptions import (
    MatchboxDeletionNotConfirmed,
    MatchboxResolutionNotFoundError,
)
from matchbox.common.factories.models import model_factory
from matchbox.common.factories.sources import source_factory


def test_get_source(api_client_and_mocks: tuple[TestClient, Mock, Mock]) -> None:
    source_testkit = source_factory(name="foo").fake_run()
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_resolution = Mock(
        return_value=source_testkit.source.to_resolution(),  # Second call (SOURCE)
    )

    response = test_client.get("/collections/default/runs/1/resolutions/foo")
    assert response.status_code == 200
    assert response.json()["resolution_type"] == "source"


def test_get_model(api_client_and_mocks: tuple[TestClient, Mock, Mock]) -> None:
    testkit = model_factory(name="test_model", description="test description")
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_resolution = Mock(
        return_value=testkit.fake_run().model.to_resolution()
    )

    response = test_client.get("/collections/default/runs/1/resolutions/test_model")

    assert response.status_code == 200
    assert response.json()["description"] == testkit.model.description


def test_get_resolution_404(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_resolution = Mock(side_effect=MatchboxResolutionNotFoundError())

    response = test_client.get("/collections/default/runs/1/resolutions/nonexistent")

    assert response.status_code == 404
    assert response.json()["entity"] == BackendResourceType.RESOLUTION


def test_insert_resolution(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Resolution metadata can be created."""
    testkit = model_factory(name="test_model").fake_run()
    test_client, mock_backend, _ = api_client_and_mocks

    response = test_client.post(
        "/collections/default/runs/1/resolutions/test_model",
        json=testkit.model.to_resolution().model_dump(),
    )

    assert response.status_code == 201
    assert (
        response.json()
        == ResourceOperationStatus(
            success=True,
            target="Resolution default/1/test_model",
            operation=CRUDOperation.CREATE,
            details=None,
        ).model_dump()
    )

    mock_backend.create_resolution.assert_called_once_with(
        resolution=testkit.model.to_resolution(),
        path=ResolutionPath(name="test_model", collection="default", run=1),
    )


def test_insert_resolution_error(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.create_resolution = Mock(side_effect=Exception("Test error"))
    testkit = model_factory()

    response = test_client.post(
        "/collections/default/runs/1/resolutions/name",
        json=testkit.fake_run().model.to_resolution().model_dump(),
    )

    assert response.status_code == 500
    assert response.json()["success"] is False
    assert response.json()["details"] == "Test error"


def test_update_resolution(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Resolution metadata can be updated."""
    testkit = model_factory(name="test_model").fake_run()
    test_client, mock_backend, _ = api_client_and_mocks

    response = test_client.put(
        "/collections/default/runs/1/resolutions/test_model",
        json=testkit.model.to_resolution().model_dump(),
    )

    assert response.status_code == 200
    assert (
        response.json()
        == ResourceOperationStatus(
            success=True,
            target="Resolution default/1/test_model",
            operation=CRUDOperation.UPDATE,
            details=None,
        ).model_dump()
    )

    mock_backend.update_resolution.assert_called_once_with(
        resolution=testkit.model.to_resolution(),
        path=ResolutionPath(name="test_model", collection="default", run=1),
    )

    # Errors are handled
    mock_backend.update_resolution = Mock(side_effect=Exception("Test error"))
    response = test_client.put(
        "/collections/default/runs/1/resolutions/name",
        json=testkit.model.to_resolution().model_dump(),
    )

    assert response.status_code == 500
    assert response.json()["success"] is False
    assert response.json()["details"] == "Test error"


# We can patch BackgroundTasks, since the api_client_and_mocks fixture
# ensures the API runs the task (not Celery)
@patch("matchbox.server.api.routers.collection.BackgroundTasks.add_task")
@patch("matchbox.server.api.routers.collection.file_to_s3")
def test_complete_upload_process(
    mock_add_task: Mock,
    mock_s3_upload: Mock,
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test the complete resolution data upload from creation through processing."""
    test_client, mock_backend, _ = api_client_and_mocks

    # Create test data
    testkit = model_factory(model_type="linker").fake_run()
    collection = testkit.resolution_path.collection
    run = testkit.resolution_path.run

    # Mock insertion of data
    mock_backend.insert_model_data = Mock(return_value=None)
    # Mock retrieval of resolution metadata and data
    mock_backend.get_resolution = Mock(return_value=testkit.model.to_resolution())
    mock_backend.get_model_data = Mock(return_value=testkit.probabilities.to_arrow())
    # Mock upload status checking - in the beginning, it's ready to receive
    mock_backend.get_resolution_stage = Mock(return_value=UploadStage.READY)

    # Step 1: Create resolution
    response = test_client.post(
        f"/collections/{collection}/runs/{run}/resolutions/{testkit.model.name}",
        json=testkit.model.to_resolution().model_dump(),
    )
    assert response.status_code == 201
    resolution_post_info = ResourceOperationStatus.model_validate(response.json())
    assert resolution_post_info.success
    assert testkit.model.name in resolution_post_info.target

    # Step 2: Upload data file
    response = test_client.post(
        f"/collections/{collection}/runs/{run}/resolutions/{testkit.model.name}/data",
        files={
            "file": (
                "results.parquet",
                table_to_buffer(testkit.probabilities.to_arrow()),
                "application/octet-stream",
            ),
        },
    )
    assert response.status_code == 202
    data_post_info = ResourceOperationStatus.model_validate(response.json())
    upload_id = data_post_info.details
    assert mock_s3_upload.called
    assert mock_add_task.called

    # Update upload status mock - now, it's complete
    mock_backend.get_resolution_stage = Mock(return_value=UploadStage.COMPLETE)

    # Step 3: Get upload status
    response = test_client.get(
        f"/collections/{collection}/runs/{run}/resolutions/{testkit.model.name}/data/status",
        params={"upload_id": upload_id},
    )
    assert response.status_code == 200
    upload_info = UploadInfo.model_validate(response.json())
    assert upload_info.stage == UploadStage.COMPLETE

    # Step 4: We can retrieve the results
    response = test_client.get(
        f"/collections/default/runs/1/resolutions/{testkit.model.name}/data"
    )
    assert response.status_code == 200
    assert response.headers["content-type"] == "application/octet-stream"


@patch("matchbox.server.api.routers.collection.BackgroundTasks.add_task")
@patch("matchbox.server.api.routers.collection.file_to_s3")
def test_set_data_404(
    mock_add_task: Mock,
    mock_s3_upload: Mock,
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test setting data for a non-existent resolution."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_resolution_stage = Mock(
        side_effect=MatchboxResolutionNotFoundError()
    )
    mock_backend.get_resolution = Mock(side_effect=MatchboxResolutionNotFoundError())

    response = test_client.post(
        "/collections/default/runs/1/resolutions/nonexistent-model/data",
        files={
            "file": (
                "hashes.parquet",
                table_to_buffer(source_factory().data_hashes),
                "application/octet-stream",
            ),
        },
    )

    # Correct error response
    assert response.status_code == 404
    error404 = NotFoundError.model_validate(response.json())
    assert error404.entity == BackendResourceType.RESOLUTION
    # Upload doesn't proceed
    mock_s3_upload.assert_not_called()
    mock_add_task.assert_not_called()


@patch("matchbox.server.api.routers.collection.BackgroundTasks.add_task")
@patch("matchbox.server.api.routers.collection.file_to_s3")
def test_set_data_file_format(
    mock_add_task: Mock,
    mock_s3_upload: Mock,
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test that file uploaded has Parquet magic bytes."""
    # Setup
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_resolution_stage = Mock(return_value=UploadStage.READY)

    # Make request with mocked background task
    response = test_client.post(
        "/collections/default/runs/1/resolutions/resolution/data",
        files={
            "file": (
                "hashes.parquet",
                b"dummy\ndata",
                "application/octet-stream",
            ),
        },
    )

    # Correct error response
    assert response.status_code == 400
    operation_status = ResourceOperationStatus.model_validate(response.json())
    assert "invalid parquet" in operation_status.details.lower()
    # Upload doesn't proceed
    mock_s3_upload.assert_not_called()
    mock_add_task.assert_not_called()


@patch("matchbox.server.api.routers.collection.BackgroundTasks.add_task")
@patch("matchbox.server.api.routers.collection.file_to_s3")
def test_set_data_already_queued(
    mock_add_task: Mock,
    mock_s3_upload: Mock,
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test attempting to upload when status is already queued."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_resolution_stage = Mock(return_value=UploadStage.PROCESSING)

    response = test_client.post(
        "/collections/default/runs/1/resolutions/resolution/data",
        files={
            "file": (
                "hashes.parquet",
                table_to_buffer(source_factory().data_hashes),
                "application/octet-stream",
            ),
        },
    )

    # Correct error response
    assert response.status_code == 400
    operation_status = ResourceOperationStatus.model_validate(response.json())
    assert "Not expecting upload" in operation_status.details
    # Upload doesn't proceed
    mock_s3_upload.assert_not_called()
    mock_add_task.assert_not_called()


def test_get_upload_status_404(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test getting upload status for a non-existent resolution."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_resolution_stage = Mock(
        side_effect=MatchboxResolutionNotFoundError()
    )

    response = test_client.get(
        "/collections/default/runs/1/resolutions/nonexistent-model/data/status"
    )

    assert response.status_code == 404
    error404 = NotFoundError.model_validate(response.json())
    error404.entity = BackendResourceType.RESOLUTION


def test_get_upload_status_gets_errors(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test getting failed upload status."""
    test_client, mock_backend, mock_tracker = api_client_and_mocks
    mock_backend.get_resolution_stage = Mock(return_value=UploadStage.READY)
    mock_tracker.set("upload_id", "error message")

    response = test_client.get(
        "/collections/default/runs/1/resolutions/nonexistent-model/data/status",
        params={"upload_id": "upload_id"},
    )

    assert response.status_code == 200
    info = UploadInfo.model_validate(response.json())
    assert info.error == "error message"


def test_get_results(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    testkit = model_factory()
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_model_data = Mock(return_value=testkit.probabilities.to_arrow())

    response = test_client.get(
        f"/collections/default/runs/1/resolutions/{testkit.model.name}/data"
    )

    assert response.status_code == 200
    assert response.headers["content-type"] == "application/octet-stream"


def test_delete_resolution(api_client_and_mocks: tuple[TestClient, Mock, Mock]) -> None:
    """Test deletion of a resolution."""
    testkit = model_factory()
    test_client, _, _ = api_client_and_mocks
    response = test_client.delete(
        f"/collections/default/runs/1/resolutions/{testkit.model.name}",
        params={"certain": True},
    )

    assert response.status_code == 200
    assert (
        response.json()
        == ResourceOperationStatus(
            success=True,
            target=f"Resolution default/1/{testkit.model.name}",
            operation=CRUDOperation.DELETE,
            details=None,
        ).model_dump()
    )


def test_delete_resolution_needs_confirmation(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test deletion of a model that requires confirmation."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.delete_resolution = Mock(
        side_effect=MatchboxDeletionNotConfirmed(children=["dedupe1", "dedupe2"])
    )

    testkit = model_factory()
    response = test_client.delete(
        f"/collections/default/runs/1/resolutions/{testkit.model.name}"
    )

    assert response.status_code == 409
    assert response.json()["success"] is False
    message = response.json()["details"]
    assert "dedupe1" in message and "dedupe2" in message


@pytest.mark.parametrize("certain", [True, False])
def test_delete_resolution_404(
    certain: bool,
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test 404 response when trying to delete a non-existent resolution."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.delete_resolution.side_effect = MatchboxResolutionNotFoundError()

    response = test_client.delete(
        "/collections/default/runs/1/resolutions/nonexistent-model",
        params={"certain": certain},
    )

    assert response.status_code == 404
    error = NotFoundError.model_validate(response.json())
    assert error.entity == BackendResourceType.RESOLUTION



"""Public evaluation helpers for Matchbox clients."""

from matchbox.client.eval.samples import (
    EvalData,
    EvaluationItem,
    ModelComparison,
    compare_models,
    create_evaluation_item,
    create_judgement,
    get_samples,
    precision_recall,
)

__all__ = [
    "EvalData",
    "EvaluationItem",
    "compare_models",
    "create_evaluation_item",
    "create_judgement",
    "get_samples",
    "ModelComparison",
    "precision_recall",
]



"""A linking methodology based on a deterministic set of conditions."""

import json

import duckdb
import polars as pl
from pydantic import Field, field_validator

from matchbox.client.models import comparison
from matchbox.client.models.linkers.base import Linker, LinkerSettings
from matchbox.common.logging import logger


class DeterministicSettings(LinkerSettings):
    """A data class to enforce the Deterministic linker's settings dictionary shape."""

    comparisons: list[str] | list[list[str]] = Field(
        description="""
            Comparison rules for matching using DuckDB SQL syntax.
            
            Can be specified as:
            - A flat list of strings: All comparisons applied in parallel (OR logic)
            - A nested list of lists: Sequential rounds of matching
            
            Flat list (parallel):
            [
                "left.company_number = right.company_number",
                "left.name = right.name",
            ]
            All comparisons applied to full datasets, results unioned.
            
            Nested list (sequential rounds):
            [
                [
                    "left.company_number = right.company_number",
                    "left.name = right.name",
                ],
                [
                    "left.name_normalised = right.name_normalised",
                    "left.website = right.website",
                ],
            ]
            Each inner list is a "round". Within each round, comparisons use OR 
            logic. After each round, matched records are removed from the pool 
            before the next round.
            
            Use left.field and right.field to refer to columns in the respective 
            sources. Supports all DuckDB SQL operations and functions.
        """,
    )

    @field_validator("comparisons", mode="before")
    @classmethod
    def validate_comparison(
        cls, value: str | list[str] | list[list[str]]
    ) -> list[list[str]]:
        """Normalise to list of lists format."""
        if isinstance(value, str):
            return [[comparison(value, dialect="duckdb")]]
        if not value:
            raise ValueError("comparisons cannot be empty")
        if all(isinstance(v, str) for v in value):
            return [[comparison(v, dialect="duckdb") for v in value]]
        if all(isinstance(v, list) for v in value):
            for round_idx, round_comparisons in enumerate(value):
                if not round_comparisons:
                    raise ValueError(f"Round {round_idx} cannot be empty")
                if not all(isinstance(c, str) for c in round_comparisons):
                    raise ValueError(f"Round {round_idx} must contain only strings")
            return [[comparison(c, dialect="duckdb") for c in r] for r in value]
        raise ValueError(
            "comparisons must be a string, list of strings, or list of lists"
        )


class DeterministicLinker(Linker):
    """A deterministic linker that links based on a set of boolean conditions.

    Uses DuckDB as the SQL backend, enabling rich SQL operations while maintaining
    a Polars DataFrame interface. Supports both parallel matching (single round)
    and sequential matching (multiple rounds where matched records are removed
    after each round).
    """

    settings: DeterministicSettings

    def prepare(self, left: pl.DataFrame, right: pl.DataFrame) -> None:
        """Prepare the linker for linking."""
        pass

    def link(self, left: pl.DataFrame, right: pl.DataFrame) -> pl.DataFrame:
        """Link the left and right dataframes.

        If comparisons is a flat list, applies all comparisons in parallel.
        If comparisons is a nested list, applies each round sequentially,
        removing matched records from the pool after each round.
        """
        con: duckdb.DuckDBPyConnection = duckdb.connect(":memory:")
        try:
            all_matches: list[pl.DataFrame] = []
            remaining_left, remaining_right = left, right

            for round_num, round_comparisons in enumerate(
                self.settings.comparisons, start=1
            ):
                if remaining_left.is_empty() or remaining_right.is_empty():
                    logger.info(f"Round {round_num}: Skipping - no records remaining")
                    break

                logger.info(
                    f"Round {round_num}: {len(remaining_left):,} left × "
                    f"{len(remaining_right):,} right"
                )

                matches = self._link_round(
                    con, remaining_left, remaining_right, round_comparisons, round_num
                )

                logger.info(f"Round {round_num}: Found {len(matches):,} matches")

                if not matches.is_empty():
                    all_matches.append(matches)
                    matched_left = matches.select("left_id").unique()
                    matched_right = matches.select("right_id").unique()
                    remaining_left = remaining_left.join(
                        matched_left,
                        left_on=self.settings.left_id,
                        right_on="left_id",
                        how="anti",
                    )
                    remaining_right = remaining_right.join(
                        matched_right,
                        left_on=self.settings.right_id,
                        right_on="right_id",
                        how="anti",
                    )

            return self._finalise_results(all_matches)
        finally:
            con.close()

    def _link_round(
        self,
        con: duckdb.DuckDBPyConnection,
        left: pl.DataFrame,
        right: pl.DataFrame,
        comparisons: list[str],
        round_num: int,
    ) -> pl.DataFrame:
        """Apply all comparisons in a round using OR logic via DuckDB."""
        con.register("left_df", left)
        con.register("right_df", right)

        subqueries: list[str] = []
        for condition in comparisons:
            subquery: str = f"""
                SELECT
                    l.{self.settings.left_id} AS left_id,
                    r.{self.settings.right_id} AS right_id,
                    1.0 AS probability
                FROM left_df l
                INNER JOIN right_df r
                    ON {condition}
            """
            subqueries.append(subquery)

        query: str = f"""
            SELECT DISTINCT *
            FROM ({" UNION ALL ".join(subqueries)})
        """

        max_est: int = self._get_max_cardinality(con, query)
        logger.info(f"Round {round_num}: Estimated max cardinality: {max_est:,}")

        return con.execute(query).pl()

    def _get_max_cardinality(self, con: duckdb.DuckDBPyConnection, query: str) -> int:
        """Get max cardinality estimate from DuckDB plan, or -1 if unavailable."""
        explain = con.execute(
            f"PRAGMA explain_output = 'all'; EXPLAIN (FORMAT json) {query}"
        ).fetchall()
        plans = {k: json.loads(v) for k, v in dict(explain).items()}

        estimates: list[dict] = []
        for plan_name, plan_type in [
            ("physical_plan", "physical"),
            ("logical_opt", "optimised"),
        ]:
            if plan := plans.get(plan_name):
                estimates.extend(self._traverse_plan(plan[0], plan_type))

        # Debug: log full tree breakdown
        logger.debug("Plan breakdown:")
        for plan_type in ["physical", "optimised"]:
            plan_ests = [e for e in estimates if e["plan_type"] == plan_type]
            if plan_ests:
                logger.debug(f"  {plan_type.capitalize()} plan:")
                for est in plan_ests:
                    indent = "    " + "  " * est["depth"]
                    logger.debug(f"{indent}{est['node_name']}: {est['cardinality']:,}")

        # Return max for info logging
        return max(
            (e["cardinality"] for e in estimates if e["cardinality"] > 0),
            default=-1,
        )

    def _traverse_plan(self, node: dict, plan_type: str, depth: int = 0) -> list[dict]:
        """Recursively collect cardinality estimates with metadata from plan tree."""
        estimates: list[dict] = []
        cardinality = node.get("extra_info", {}).get("Estimated Cardinality")
        if cardinality is not None:
            estimates.append(
                {
                    "cardinality": int(cardinality),
                    "node_name": node.get("name", "UNKNOWN"),
                    "depth": depth,
                    "plan_type": plan_type,
                }
            )
        for child in node.get("children", []):
            estimates.extend(self._traverse_plan(child, plan_type, depth + 1))
        return estimates

    def _finalise_results(self, all_matches: list[pl.DataFrame]) -> pl.DataFrame:
        """Combine matches from all rounds and ensure correct schema."""
        if all_matches:
            return pl.concat(all_matches).with_columns(
                pl.col("probability").cast(pl.Float32)
            )
        return pl.DataFrame(
            {"left_id": [], "right_id": [], "probability": []}
        ).with_columns(pl.col("probability").cast(pl.Float32))



"""Definition of model inputs."""

from enum import StrEnum
from typing import TYPE_CHECKING, Any, Self

import duckdb
import polars as pl
from polars import DataFrame as PolarsDataFrame
from sqlglot import expressions, parse_one
from sqlglot import select as sqlglot_select

from matchbox.client import _handler
from matchbox.client.models.dedupers.base import Deduper, DeduperSettings
from matchbox.client.models.linkers.base import Linker, LinkerSettings
from matchbox.common.db import QueryReturnClass, QueryReturnType
from matchbox.common.dtos import (
    QueryCombineType,
    QueryConfig,
)
from matchbox.common.transform import truth_float_to_int, truth_int_to_float

if TYPE_CHECKING:
    from matchbox.client.dags import DAG
    from matchbox.client.models import Model
    from matchbox.client.sources import Source
else:
    DAG = Any
    Model = Any
    Source = Any


class CacheMode(StrEnum):
    """Settings determining what query data gets cached."""

    OFF = "off"
    RAW = "raw"
    CLEAN = "clean"


class Query:
    """Queriable input to a model."""

    def __init__(
        self,
        *sources: Source,
        dag: DAG,
        model: Model | None = None,
        combine_type: QueryCombineType = QueryCombineType.CONCAT,
        threshold: float | None = None,
        cleaning: dict[str, str] | None = None,
    ) -> None:
        """Initialise query.

        Args:
            sources: List of sources to query from
            dag: DAG containing sources and models.
            model (optional): Model to use to resolve sources. It can only be missing
                if querying from a single source.
            combine_type (optional): How to combine the data from different sources.
                Default is `concat`.

                * If `concat`, concatenate all sources queried without any merging.
                    Multiple rows per ID, with null values where data isn't available
                * If `explode`, outer join on Matchbox ID. Multiple rows per ID,
                    with one for every unique combination of data requested
                    across all sources
                * If `set_agg`, join on Matchbox ID, group on Matchbox ID, then
                    aggregate to nested lists of unique values. One row per ID,
                    but all requested data is in nested arrays

            threshold (optional): The threshold to use for creating clusters
                If None, uses the resolutions' default threshold
                If an integer, uses that threshold for the specified resolution, and the
                resolution's cached thresholds for its ancestors

            cleaning (optional): A dictionary mapping an output column name to a SQL
                expression that will populate a new column.
        """
        self.raw_data: PolarsDataFrame | None = None
        self.data: PolarsDataFrame | None = None
        self.dag = dag
        self.sources = sources
        self.model = model
        self.combine_type = combine_type
        self.threshold = threshold
        self.cleaning = cleaning
        self._cache_mode: CacheMode = CacheMode.OFF

    @property
    def config(self) -> QueryConfig:
        """The query configuration for the current DAG."""
        return QueryConfig(
            source_resolutions=[source.name for source in self.sources],
            model_resolution=self.model.name if self.model else None,
            combine_type=self.combine_type,
            threshold=truth_float_to_int(self.threshold) if self.threshold else None,
            cleaning=self.cleaning,
        )

    @classmethod
    def from_config(cls, config: QueryConfig, dag: DAG) -> Self:
        """Create query from config.

        The DAG must have had relevant sources and model added already.

        Args:
            config: The QueryConfig to reconstruct from.
            dag: The DAG containing the sources and model.

        Returns:
            A reconstructed Query instance.
        """
        # Get sources from DAG
        sources = [dag.get_source(res) for res in config.source_resolutions]

        # Get model if specified
        model = (
            dag.get_model(config.model_resolution) if config.model_resolution else None
        )

        # Convert threshold back to float
        threshold = truth_int_to_float(config.threshold) if config.threshold else None

        return cls(
            *sources,
            dag=dag,
            model=model,
            combine_type=config.combine_type,
            threshold=threshold,
            cleaning=config.cleaning,
        )

    def set_cache_mode(self, mode: CacheMode = CacheMode.OFF) -> Self:
        """Configures caching behaviour of query operations.

        * If "off" (default), doesn't cache anything
        * If "raw", caches data as fetched from the source
        * If "clean", it additionally caches the result of applying the
            cleaning dict.
        """
        self._cache_mode = mode
        return self

    def _set_cache(
        self, raw_data: PolarsDataFrame | None, data: PolarsDataFrame | None
    ) -> None:
        self.raw_data = None
        self.data = None

        if self._cache_mode in [CacheMode.RAW, CacheMode.CLEAN]:
            self.raw_data = raw_data

        if self._cache_mode == CacheMode.CLEAN:
            self.data = data

    def run(
        self,
        return_type: QueryReturnType = QueryReturnType.POLARS,
        return_leaf_id: bool = False,
        batch_size: int | None = None,
        reuse_cache: bool = False,
    ) -> QueryReturnClass:
        """Runs queries against the selected backend.

        Args:
            return_type (optional): Type of dataframe returned, defaults to "polars".
                Other options are "pandas" and "arrow".
            return_leaf_id (optional): Whether matchbox IDs for source clusters should
                be saved as a byproduct in the `leaf_ids` attribute.
            batch_size (optional): The size of each batch when fetching data from the
                warehouse, which helps reduce memory usage and load on the database.
                Default is None.
            reuse_cache: Whether to re-use raw cached data if available.

        Returns: Data in the requested return type

        Raises:
            MatchboxEmptyServerResponse: If no data was returned by the server.
        """
        if reuse_cache:
            if self.data is not None:
                result = _convert_df(self.data, return_type=return_type)
                self._set_cache(self.raw_data, self.data)
                return result
            elif self.raw_data is not None:
                result = self.clean(self.cleaning, return_type=return_type)
                self._set_cache(self.raw_data, self.data)
                return result

        source_results: list[PolarsDataFrame] = []
        for source in self.sources:
            mb_ids = pl.from_arrow(
                _handler.query(
                    source=source.resolution_path,
                    resolution=self.model.resolution_path if self.model else None,
                    threshold=self.config.threshold,
                    return_leaf_id=return_leaf_id,
                )
            )

            raw_batches = source.fetch(
                qualify_names=True,
                batch_size=batch_size,
                return_type=QueryReturnType.POLARS,
            )

            processed_batches = [
                b.join(
                    other=mb_ids,
                    left_on=source.qualified_key,
                    right_on="key",
                    how="inner",
                )
                for b in raw_batches
            ]
            source_results.append(pl.concat(processed_batches, how="vertical"))

        # Process all data and return a single result
        tables: list[PolarsDataFrame] = list(source_results)

        # Combine results based on combine_type
        if return_leaf_id:
            concatenated = pl.concat(tables, how="diagonal")
            self.leaf_id = concatenated.select(["id", "leaf_id"])

        if self.config.combine_type == QueryCombineType.CONCAT:
            if return_leaf_id:  # can reuse the concatenated dataframe
                raw_data = concatenated.drop(["leaf_id"])
            else:
                raw_data = pl.concat(tables, how="diagonal")
        else:
            raw_data = tables[0].drop("leaf_id", strict=False)
            for table in tables[1:]:
                raw_data = raw_data.join(
                    table.drop("leaf_id", strict=False),
                    on="id",
                    how="full",
                    coalesce=True,
                )

            raw_data = raw_data.select(["id", pl.all().exclude("id")])

            if self.config.combine_type == QueryCombineType.SET_AGG:
                # Aggregate into lists
                agg_expressions = [
                    pl.col(col).unique() for col in raw_data.columns if col != "id"
                ]
                raw_data = raw_data.group_by("id").agg(agg_expressions)

        clean_data = _clean(data=raw_data, cleaning_dict=self.config.cleaning)

        self._set_cache(raw_data, clean_data)
        return _convert_df(clean_data, return_type=return_type)

    def clean(
        self,
        cleaning: dict[str, str] | None,
        return_type: QueryReturnType = QueryReturnType.POLARS,
    ) -> QueryReturnClass:
        """Change cleaning dictionary and re-apply cleaning, if raw data was cached.

        Args:
            cleaning: A dictionary mapping field aliases to SQL expressions.
                The SQL expressions can reference columns in the data using their names.
                If None, no cleaning is applied and the original data is returned.
                `SourceConfig.f()` can be used to help reference qualified fields.
            return_type (optional): Type of dataframe returned, defaults to "polars".
                    Other options are "pandas" and "arrow".
        """
        if self.raw_data is None:
            raise RuntimeError("No raw data is stored in this query.")

        clean_data = _convert_df(
            data=_clean(
                data=self.raw_data,
                cleaning_dict=cleaning,
            ),
            return_type=return_type,
        )

        self.cleaning = cleaning

        self._set_cache(self.raw_data, clean_data)

        return clean_data

    def deduper(
        self,
        name: str,
        model_class: Deduper,
        model_settings: DeduperSettings,
        description: str | None = None,
    ) -> Model:
        """Create deduper for data in this query."""
        return self.dag.model(
            name=name,
            description=description,
            model_class=model_class,
            model_settings=model_settings,
            left_query=self,
        )

    def linker(
        self,
        other_query: Self,
        name: str,
        model_class: Linker,
        model_settings: LinkerSettings,
        description: str | None = None,
    ) -> Model:
        """Create linker for data in this query and another query."""
        return self.dag.model(
            name=name,
            description=description,
            model_class=model_class,
            model_settings=model_settings,
            left_query=self,
            right_query=other_query,
        )


def _clean(
    data: pl.DataFrame,
    cleaning_dict: dict[str, str] | None,
) -> pl.DataFrame:
    """Clean data using DuckDB with the provided cleaning SQL.

    * ID is passed through automatically
    * If present, leaf_id is passed through automatically
    * Columns not mentioned in the cleaning_dict are dropped
    * Each key in cleaning_dict is an alias for a SQL expression

    Args:
        data: Raw polars dataframe to clean
        cleaning_dict: A dictionary mapping field aliases to SQL expressions.
            The SQL expressions can reference columns in the data using their names.
            If None, no cleaning is applied and the original data is returned.
            `SourceConfig.f()` can be used to help reference qualified fields.

    Returns:
        Cleaned polars dataframe
    """
    if cleaning_dict is None:
        return data

    def _add_column(name: str) -> expressions.Alias:
        """Select and alias a column with the given name."""
        return expressions.Alias(
            this=expressions.Column(
                this=expressions.Identifier(this=name, quoted=False)
            ),
            alias=expressions.Identifier(this=name, quoted=False),
        )

    # Always select 'id'
    to_select: list[expressions.Expression] = [_add_column("id")]

    # Add leaf_id if it exists
    if "leaf_id" in data.columns:
        to_select.append(_add_column("leaf_id"))

    # Parse and add each SQL expression from cleaning_dict
    for alias, sql in cleaning_dict.items():
        stmt = parse_one(sql, dialect="duckdb")
        to_select.append(expressions.alias_(stmt, alias))

    query = sqlglot_select(*to_select, dialect="duckdb").from_("data")

    with duckdb.connect(":memory:") as conn:
        conn.register("data", data)
        return conn.execute(query.sql(dialect="duckdb")).pl()


def _convert_df(
    data: PolarsDataFrame, return_type: QueryReturnType
) -> QueryReturnClass:
    match return_type:
        case QueryReturnType.POLARS:
            return data
        case QueryReturnType.PANDAS:
            return data.to_pandas()
        case QueryReturnType.ARROW:
            return data.to_arrow()
        case _:
            raise ValueError(f"Return type {return_type} is invalid")



"""Main CLI entry point for Matchbox."""

import typer
from rich import print

from matchbox.client import _handler
from matchbox.client.cli.eval.run import evaluate

app = typer.Typer(
    name="matchbox", help="Matchbox: Entity resolution and data linking framework"
)


@app.command()
def health() -> None:
    """Checks the health of the Matchbox server."""
    response = _handler.healthcheck()
    print(response)


app.command(name="eval")(evaluate)

if __name__ == "__main__":
    app()



"""Evaluation logic for PostgreSQL adapter."""

from datetime import UTC, datetime, timedelta

import numpy as np
import polars as pl
import pyarrow as pa
from pyarrow import Table
from sqlalchemy import BIGINT, func, select

from matchbox.common.arrow import (
    SCHEMA_CLUSTER_EXPANSION,
    SCHEMA_EVAL_SAMPLES,
    SCHEMA_JUDGEMENTS,
)
from matchbox.common.db import sql_to_df
from matchbox.common.dtos import ModelResolutionPath
from matchbox.common.eval import Judgement, ModelComparison, precision_recall
from matchbox.common.exceptions import (
    MatchboxTooManySamplesRequested,
    MatchboxUserNotFoundError,
)
from matchbox.common.transform import hash_cluster_leaves
from matchbox.server.postgresql.db import MBDB
from matchbox.server.postgresql.orm import (
    Clusters,
    ClusterSourceKey,
    Contains,
    EvalJudgements,
    PKSpace,
    Probabilities,
    Resolutions,
    SourceConfigs,
    Users,
)
from matchbox.server.postgresql.utils.db import (
    compile_sql,
    ingest_to_temporary_table,
)
from matchbox.server.postgresql.utils.query import build_unified_query


def insert_judgement(judgement: Judgement) -> None:
    """Record judgement to server."""
    # Note: we don't currently check that the shown cluster ID points to
    # the source cluster IDs. We must assume this is well-formed.

    # Check that the user exists
    with MBDB.get_session() as session:
        if not session.scalar(
            select(Users.name).where(Users.user_id == judgement.user_id)
        ):
            raise MatchboxUserNotFoundError(user_id=judgement.user_id)

    for leaves in judgement.endorsed:
        with MBDB.get_session() as session:
            # Compute hash corresponding to set of source clusters (leaves)
            leaf_hashes = [
                session.scalar(
                    select(Clusters.cluster_hash).where(Clusters.cluster_id == leaf_id)
                )
                for leaf_id in leaves
            ]
            endorsed_cluster_hash = hash_cluster_leaves(leaf_hashes)

            # If cluster with this hash does not exist, create it.
            # Note that only endorsed clusters might be new. The cluster shown to
            # the user is guaranteed to exist in the backend; we have checked above.
            if not (
                endorsed_cluster_id := session.scalar(
                    select(Clusters.cluster_id).where(
                        Clusters.cluster_hash == endorsed_cluster_hash
                    )
                )
            ):
                endorsed_cluster_id = PKSpace.reserve_block(
                    table="clusters", block_size=1
                )
                session.add(
                    Clusters(
                        cluster_id=endorsed_cluster_id,
                        cluster_hash=endorsed_cluster_hash,
                    )
                )
                for leaf_id in leaves:
                    session.add(Contains(root=endorsed_cluster_id, leaf=leaf_id))

            session.add(
                EvalJudgements(
                    user_id=judgement.user_id,
                    shown_cluster_id=judgement.shown,
                    endorsed_cluster_id=endorsed_cluster_id,
                    timestamp=datetime.now(UTC),
                )
            )

            session.commit()


def get_judgements() -> tuple[pa.Table, pa.Table]:
    """Get all judgements from server."""

    def _cast_tables(
        judgements: pl.DataFrame, cluster_expansion: pl.DataFrame
    ) -> tuple[pa.Table, pa.Table]:
        """Cast judgement tables to conform to data transfer schema."""
        judgements = judgements.cast(pl.Schema(SCHEMA_JUDGEMENTS))
        cluster_expansion = cluster_expansion.cast(pl.Schema(SCHEMA_CLUSTER_EXPANSION))

        return (
            judgements.to_arrow().cast(SCHEMA_JUDGEMENTS),
            cluster_expansion.to_arrow().cast(SCHEMA_CLUSTER_EXPANSION),
        )

    judgements_stmt = select(
        EvalJudgements.user_id,
        EvalJudgements.endorsed_cluster_id.label("endorsed"),
        EvalJudgements.shown_cluster_id.label("shown"),
    )

    with MBDB.get_adbc_connection() as conn:
        judgements = sql_to_df(
            stmt=compile_sql(judgements_stmt),
            connection=conn.dbapi_connection,
            return_type="polars",
        )

    if not len(judgements):
        cluster_expansion = pl.DataFrame(schema=pl.Schema(SCHEMA_CLUSTER_EXPANSION))
        return _cast_tables(judgements, cluster_expansion)

    shown_clusters = set(judgements["shown"].to_list())
    endorsed_clusters = set(judgements["endorsed"].to_list())
    referenced_clusters = Table.from_pydict(
        {"root": list(shown_clusters | endorsed_clusters)}
    )

    with ingest_to_temporary_table(
        table_name="judgements",
        schema_name="mb",
        column_types={
            "root": BIGINT,
        },
        data=referenced_clusters,
    ) as temp_table:
        cluster_expansion_stmt = (
            select(temp_table.c.root, func.array_agg(Contains.leaf).label("leaves"))
            .select_from(temp_table)
            .join(Contains, Contains.root == temp_table.c.root)
            .group_by(temp_table.c.root)
        )

        with MBDB.get_adbc_connection() as conn:
            cluster_expansion = sql_to_df(
                stmt=compile_sql(cluster_expansion_stmt),
                connection=conn.dbapi_connection,
                return_type="polars",
            )

    return _cast_tables(judgements, cluster_expansion)


def sample(n: int, resolution_path: ModelResolutionPath, user_id: int) -> Table:
    """Sample some clusters from a resolution."""
    # Not currently checking validity of the user_id
    # If the user ID does not exist, the exclusion by previous judgements breaks
    if n > 100:
        # This reasonable assumption means simple "IS IN" function later is fine
        raise MatchboxTooManySamplesRequested("Can only sample 100 entries at a time.")

    with MBDB.get_session() as session:
        # Use ORM to get resolution metadata
        resolution_orm = Resolutions.from_path(path=resolution_path, session=session)
        resolution_id = resolution_orm.resolution_id
        truth = resolution_orm.truth

    # Get a list of cluster IDs and features for this resolution and user
    user_judgements = (
        select(EvalJudgements).where(EvalJudgements.user_id == user_id).subquery()
    )
    cluster_features_stmt = (
        select(
            Probabilities.cluster_id,
            # We expect only one probability per cluster within one resolution
            func.max(Probabilities.probability).label("probability"),
            func.max(user_judgements.c.timestamp).label("latest_ts"),
        )
        .join(
            user_judgements,
            Probabilities.cluster_id == user_judgements.c.shown_cluster_id,
            isouter=True,
        )
        .where(
            Probabilities.resolution_id == resolution_id,
        )
        .group_by(Probabilities.cluster_id)
    )

    with MBDB.get_adbc_connection() as conn:
        cluster_features = sql_to_df(
            stmt=compile_sql(cluster_features_stmt),
            connection=conn.dbapi_connection,
            return_type="polars",
        )

    # Exclude clusters recently judged by this user
    to_sample = cluster_features.filter(
        (pl.col("latest_ts") < datetime.now(UTC) - timedelta(days=365))
        | (pl.col("latest_ts").is_null())
    )

    # Return early if nothing to sample from
    if not len(to_sample):
        return pl.DataFrame(schema=pl.Schema(SCHEMA_EVAL_SAMPLES)).to_arrow()

    # Sample proportionally to distance from the truth, and get 1D array
    distances = np.abs(to_sample.select("probability").to_numpy() - truth)[:, 0]
    # Add small noise to avoid division by 0 if all distances are 0
    unnormalised_probs = distances + 0.001
    probs = unnormalised_probs / unnormalised_probs.sum()

    # With fewer clusters than requested, return all
    if to_sample.shape[0] <= n:
        sampled_cluster_ids = to_sample.select("cluster_id").to_series().to_list()
    else:
        indices = np.random.choice(to_sample.shape[0], size=n, p=probs, replace=False)
        sampled_cluster_ids = (
            to_sample[indices].select("cluster_id").to_series().to_list()
        )

    # Get all info we need for the cluster IDs we've sampled, i.e.:
    # source cluster IDs, keys and source resolutions
    with MBDB.get_adbc_connection() as conn:
        source_clusters = (
            select(Contains.root, Contains.leaf)
            .where(Contains.root.in_(sampled_cluster_ids))
            .subquery()
        )
        enrich_stmt = (
            select(
                source_clusters.c.root,
                source_clusters.c.leaf,
                ClusterSourceKey.key,
                Resolutions.name.label("source"),
            )
            .select_from(source_clusters)
            .join(
                ClusterSourceKey,
                ClusterSourceKey.cluster_id == source_clusters.c.leaf,
            )
            .join(
                SourceConfigs,
                SourceConfigs.source_config_id == ClusterSourceKey.source_config_id,
            )
            .join(
                Resolutions,
                Resolutions.resolution_id == SourceConfigs.resolution_id,
            )
        )

        final_samples = sql_to_df(
            stmt=compile_sql(enrich_stmt),
            connection=conn.dbapi_connection,
            return_type="polars",
        )
        return final_samples.cast(pl.Schema(SCHEMA_EVAL_SAMPLES)).to_arrow()


def compare_models(
    resolutions: list[ModelResolutionPath],
    judgements: pl.DataFrame,
    expansion: pl.DataFrame,
) -> ModelComparison:
    """Compare models on the basis of precision and recall."""

    def get_root_leaf(resolution_path: ModelResolutionPath) -> Table:
        with MBDB.get_session() as session:
            resolution = Resolutions.from_path(resolution_path, session=session)
            # The session which fetched the resolution needs to be alive while
            # the root-leaf query is built
            root_leaf_query = build_unified_query(
                resolution, threshold=resolution.truth
            )

        with MBDB.get_adbc_connection() as conn:
            root_leaf_results = sql_to_df(
                stmt=compile_sql(root_leaf_query),
                connection=conn.dbapi_connection,
                return_type="arrow",
            )
            return root_leaf_results.rename_columns(
                {"root_id": "root", "leaf_id": "leaf"}
            ).select(["root", "leaf"])

    models_root_leaf = [pl.from_arrow(get_root_leaf(res)) for res in resolutions]

    pr_values = precision_recall(
        models_root_leaf=models_root_leaf,
        judgements=judgements,
        expansion=expansion,
    )

    return dict(zip(resolutions, pr_values, strict=True))
